# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time
from typing import Type

from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field


class WaitForSecondsToolInput(BaseModel):
    """Input for the WaitForSecondsTool tool."""

    seconds: int = Field(..., description="The number of seconds to wait")


class WaitForSecondsTool(BaseTool):
    """Wait for a specified number of seconds"""

    name: str = "WaitForSecondsTool"
    description: str = (
        "A tool for waiting. "
        "Useful for pausing execution for a specified number of seconds. "
        "Input should be the number of seconds to wait."
        "Maximum allowed time is 10 seconds"
    )

    args_schema: Type[WaitForSecondsToolInput] = WaitForSecondsToolInput

    def _run(self, seconds: int):
        """Waits for the specified number of seconds."""
        if seconds > 10:
            seconds = 10
        time.sleep(seconds)
        return f"Waited for {seconds} seconds."



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import time
from typing import List, Optional, Type, cast

import cv2
import numpy as np
from geometry_msgs.msg import Point, PoseStamped, Quaternion, TransformStamped
from langchain_core.tools import BaseTool
from nav2_msgs.action import NavigateToPose
from nav_msgs.msg import OccupancyGrid
from pydantic import BaseModel, Field
from rclpy.action import ActionClient
from tf_transformations import euler_from_quaternion, quaternion_from_euler

from rai.communication.ros2 import ROS2Message
from rai.communication.ros2.connectors import ROS2Connector
from rai.messages import MultimodalArtifact
from rai.tools.ros2.base import BaseROS2Tool, BaseROS2Toolkit

action_client: Optional[ActionClient] = None
current_action_id: Optional[str] = None
current_feedback: Optional[NavigateToPose.Feedback] = None
current_result: Optional[NavigateToPose.Result] = None


class Nav2Toolkit(BaseROS2Toolkit):
    connector: ROS2Connector
    frame_id: str = Field(
        default="map", description="The frame id of the Nav2 stack (map, odom, etc.)"
    )
    action_name: str = Field(
        default="navigate_to_pose", description="The name of the NavigateToPose action"
    )

    def get_tools(self) -> List[BaseTool]:
        return [
            NavigateToPoseTool(
                connector=self.connector,
                frame_id=self.frame_id,
                action_name=self.action_name,
            ),
            CancelNavigateToPoseTool(connector=self.connector),
            GetNavigateToPoseFeedbackTool(connector=self.connector),
            GetNavigateToPoseResultTool(connector=self.connector),
        ]


class NavigateToPoseToolInput(BaseModel):
    x: float = Field(..., description="The x coordinate of the pose")
    y: float = Field(..., description="The y coordinate of the pose")
    z: float = Field(..., description="The z coordinate of the pose")
    yaw: float = Field(..., description="The yaw angle of the pose")


class NavigateToPoseTool(BaseROS2Tool):
    name: str = "navigate_to_pose"
    description: str = "Navigate to a specific pose"

    args_schema: Type[NavigateToPoseToolInput] = NavigateToPoseToolInput

    frame_id: str = Field(
        default="map", description="The frame id of the Nav2 stack (map, odom, etc.)"
    )
    action_name: str = Field(
        default="navigate_to_pose", description="The name of the NavigateToPose action"
    )

    def on_feedback(self, feedback: NavigateToPose.Feedback) -> None:
        global current_feedback
        current_feedback = feedback

    def on_done(self, result: NavigateToPose.Result) -> None:
        global current_result
        current_result = result

    def _run(self, x: float, y: float, z: float, yaw: float) -> str:
        global action_client
        if action_client is None:
            action_client = ActionClient(
                self.connector.node, NavigateToPose, self.action_name
            )

        pose = PoseStamped()
        pose.header.frame_id = self.frame_id
        pose.header.stamp = self.connector.node.get_clock().now().to_msg()
        pose.pose.position.x = x
        pose.pose.position.y = y
        pose.pose.position.z = z
        quat = quaternion_from_euler(0, 0, yaw)
        pose.pose.orientation = Quaternion(x=quat[0], y=quat[1], z=quat[2], w=quat[3])

        goal = {
            "pose": {
                "header": {
                    "frame_id": self.frame_id,
                    "stamp": self.connector.node.get_clock().now().to_msg(),
                },
                "pose": {
                    "position": {"x": x, "y": y, "z": z},
                    "orientation": {
                        "x": quat[0],
                        "y": quat[1],
                        "z": quat[2],
                        "w": quat[3],
                    },
                },
            }
        }

        msg = ROS2Message(payload=goal)
        action_id = self.connector.start_action(
            action_data=msg,
            target=self.action_name,
            msg_type="nav2_msgs/action/NavigateToPose",
            on_feedback=self.on_feedback,
            on_done=self.on_done,
        )
        global current_action_id
        current_action_id = action_id

        return "Navigating to pose"


class GetNavigateToPoseFeedbackTool(BaseROS2Tool):
    name: str = "get_navigate_to_pose_feedback"
    description: str = "Get the feedback of the navigate to pose action"

    def _run(self) -> str:
        global current_feedback
        return str(current_feedback)


class GetNavigateToPoseResultTool(BaseROS2Tool):
    name: str = "get_navigate_to_pose_result"
    description: str = "Get the result of the navigate to pose action"

    def _run(self) -> str:
        global current_result
        if current_result is None:
            return "Action is not done yet"
        return str(current_result.result().result)


class CancelNavigateToPoseTool(BaseROS2Tool):
    name: str = "cancel_navigate_to_pose"
    description: str = "Cancel the navigate to pose action"

    def _run(self) -> str:
        global current_action_id
        if current_action_id is None:
            return "No action to cancel"
        self.connector.terminate_action(current_action_id)
        return "Action cancelled"


class GetOccupancyGridTool(BaseROS2Tool):
    """Get the current map as an image with the robot's position marked on it (red dot)."""

    name: str = "GetOccupancyGridTool"
    description: str = "A tool for getting the current map as an image with the robot's position marked on it."

    response_format: str = "content_and_artifact"
    image_width: int = 1500
    debug: bool = False

    def _postprocess_msg(self, map_msg: OccupancyGrid, transform: TransformStamped):
        width = cast(int, map_msg.info.width)
        height = cast(int, map_msg.info.height)
        resolution = cast(float, map_msg.info.resolution)
        origin_position = cast(Point, map_msg.info.origin.position)
        origin_orientation = cast(Quaternion, map_msg.info.origin.orientation)

        data = np.array(map_msg.data).reshape((height, width))

        # Convert the OccupancyGrid values to grayscale image (0-255)
        # the final image shape is (self.image_width, self.image_width), scale to fit
        scale = self.image_width / max(width, height)
        width = int(width * scale)
        height = int(height * scale)
        data = cv2.resize(data, (width, height), interpolation=cv2.INTER_NEAREST)
        resolution = resolution / scale
        image = np.zeros_like(data, dtype=np.uint8)
        image[data == -1] = 127  # Unknown space
        image[data == 0] = 255  # Free space
        image[data > 0] = 0  # Occupied space

        # Calculate robot's position in the image
        robot_x = cast(
            float, (transform.transform.translation.x - origin_position.x) / resolution
        )

        robot_y = cast(
            float, (transform.transform.translation.y - origin_position.y) / resolution
        )

        _, _, yaw = euler_from_quaternion(
            [
                origin_orientation.x,  # type: ignore
                origin_orientation.y,  # type: ignore
                origin_orientation.z,  # type: ignore
                origin_orientation.w,  # type: ignore
            ]
        )
        # Rotate the robot's position based on the yaw angle
        rotated_x = robot_x * np.cos(yaw) - robot_y * np.sin(yaw)
        rotated_y = robot_x * np.sin(yaw) + robot_y * np.cos(yaw)
        robot_x = int(rotated_x)
        robot_y = int(rotated_y)

        image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)

        # Draw the robot's position as an arrow
        if 0 <= robot_x < width and 0 <= robot_y < height:
            _, _, yaw = euler_from_quaternion(
                [
                    transform.transform.rotation.x,  # type: ignore
                    transform.transform.rotation.y,  # type: ignore
                    transform.transform.rotation.z,  # type: ignore
                    transform.transform.rotation.w,  # type: ignore
                ]
            )
            arrow_length = 100
            arrow_end_x = int(robot_x + arrow_length * np.cos(yaw))
            arrow_end_y = int(robot_y + arrow_length * np.sin(yaw))
            cv2.arrowedLine(
                image, (robot_x, robot_y), (arrow_end_x, arrow_end_y), (0, 0, 255), 5
            )

        image = cv2.flip(image, 1)

        step_size_m: float = 2.0  # Step size for grid lines in meters, adjust as needed
        step_size_pixels = int(step_size_m / resolution)
        # print(step_size_pixels, scale)
        for x in range(0, width, step_size_pixels):
            cv2.line(
                img=image,
                pt1=(x, 40),
                pt2=(x, height),
                color=(200, 200, 200),
                thickness=1,
            )
            cv2.putText(
                img=image,
                text=f"{x * resolution + origin_position.x:.1f}",
                org=(x, 30),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=1.5,
                color=(0, 0, 255),
                thickness=2,
                lineType=cv2.LINE_AA,
            )
        for y in range(0, height, step_size_pixels):
            cv2.line(
                img=image,
                pt1=(0, y),
                pt2=(width, y),
                color=(200, 200, 200),
                thickness=1,
            )
            cv2.putText(
                img=image,
                text=f"{y * resolution + origin_position.y:.1f}",
                org=(15, y + 35),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=1.5,
                color=(0, 0, 255),
                thickness=2,
                lineType=cv2.LINE_AA,
            )
        # Encode into PNG base64
        _, buffer = cv2.imencode(".png", image)

        if self.debug:
            cv2.imwrite(f"map{time.time()}.png", image)
        return base64.b64encode(buffer.tobytes()).decode("utf-8")

    def _run(self):
        """Gets the current map from the specified topic."""
        map_msg = self.connector.receive_message("/map", timeout_sec=10).payload
        transform = self.connector.get_transform(
            target_frame="map", source_frame="base_link", timeout_sec=10
        )

        if map_msg is None or transform is None:
            return {"content": "Failed to get the map, wrong topic?"}

        base64_image = self._postprocess_msg(map_msg, transform)
        return "Map grabbed successfully", MultimodalArtifact(
            images=[base64_image], audios=[]
        )



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from dataclasses import dataclass
from typing import Any, Callable, Optional

import numpy as np
import sounddevice as sd
from numpy._typing import NDArray
from pydub import AudioSegment
from scipy.signal import resample


class SoundDeviceError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


@dataclass
class SoundDeviceConfig:
    """
    Configuration settings for a sound device.

    This dataclass holds configuration parameters for audio input/output devices.
    It ensures that at least one identifier (`device_number` or `device_name`) is
    provided when initialized.

    Parameters
    ----------
    stream : bool, optional
        Whether the device should operate in streaming mode. Default is False.
    block_size : int, optional
        The block size for audio processing. Default is 1024.
    dtype : str, optional
        The data type of the audio stream (e.g., "int16", "float32"). Default is "int16".
    channels : int, optional
        The number of audio channels (e.g., 1 for mono, 2 for stereo). Default is 1.
    consumer_sampling_rate : Optional[int], optional
        The desired sampling rate for the audio consumer. Default is None.
    device_number : Optional[int], optional
        The device number for the sound device. If None, `device_name` must be set. Default is None.
    device_name : Optional[str], optional
        The name of the sound device. If None, `device_number` must be set. Default is None.
    is_input : bool, optional
        Indicates whether the device is used for input (recording). Default is False.
    is_output : bool, optional
        Indicates whether the device is used for output (playback). Default is False.

    Raises
    ------
    ValueError
        If neither `device_number` nor `device_name` is provided.
    """

    stream: bool = False
    block_size: int = 1024
    dtype: str = "int16"
    channels: Optional[int] = None
    consumer_sampling_rate: Optional[int] = None
    device_number: Optional[int] = None
    device_name: Optional[str] = None
    is_input: bool = False
    is_output: bool = False

    def __post_init__(self):
        if self.device_number is None and self.device_name is None:
            raise ValueError("Either 'device_number' or 'device_name' must be set.")
        if not self.is_input and not self.is_output:
            raise ValueError("Either 'is_input' or 'is_output' must be True.")


class SoundDeviceAPI:
    def __init__(self, config: SoundDeviceConfig):
        self.device_name = ""

        self.config = config
        sd.default.latency = ("low", "low")  # type: ignore
        if config.device_name:
            self.device_name = config.device_name
            devices = sd.query_devices()
            devices = list(devices) if isinstance(devices, sd.DeviceList) else [devices]
            for device in devices:
                if device["name"] == config.device_name:  # type: ignore
                    self.device_number = int(device["index"])  # type: ignore
                    break
        else:
            self.device_number = config.device_number
        try:
            device_data = sd.query_devices(device=self.device_number)
        except AttributeError:
            raise SoundDeviceError(
                f"Device {self.device_name} was not found for configuration"
            )
        self.sample_rate = int(device_data["default_samplerate"])  # type: ignore
        if self.config.channels is None:
            self.in_channels = int(device_data["max_input_channels"])  # type: ignore
            self.out_channels = int(device_data["max_output_channels"])  # type: ignore
        else:
            self.in_channels = self.config.channels
            self.out_channels = self.config.channels

        self.read_flag = config.is_input
        self.write_flag = config.is_output
        self.stream_flag = config.stream
        self.in_stream = None
        self.out_stream = None

    def write(self, data: AudioSegment, blocking: bool = False, loop: bool = False):
        """
        Write data to the sound device.

        Parameters
        ----------
        data : NDArray
            Data to be written.
        blocking : bool, optional
            If True, the function will block until the sound is played. Defaults to False.
        loop : bool, optional
            If True, the data will loop continuously. Defaults to False.

        Notes
        -----
        - If `blocking` is True, the function will block until the sound is played.
        - If both `blocking` and `loop` are True, the function will block indefinitely.
        - Calling this function will stop any sound that is currently playing and any
          recording currently happening.
        - Call `stop()` or `read()` to stop the sound.
        """
        if not self.write_flag:
            raise SoundDeviceError(f"{self.device_name} does not support writing!")

        audio = np.array(data.get_array_of_samples())
        sd.play(
            audio,
            samplerate=data.frame_rate,
            blocking=blocking,
            loop=loop,
            device=self.device_number,
        )

    def read(self, time: float, blocking: bool = False) -> AudioSegment:
        """
        Read data from the sound device.

        Parameters
        ----------
        time : float
            Time in seconds to read.
        blocking : bool, optional
            If True, the function will block until the sound is read. If False,
            the function will return immediately with an unpopulated ndarray.
            Defaults to False.

        Returns
        -------
        NDArray
            Data read from the sound device.

        Notes
        -----
        - If `blocking` is True, the function will block until the sound is read.
        - If `blocking` is False, the function will return immediately with an
          unpopulated ndarray. The array will be populated with the sound data
          after the read is complete.
        - Calling this function will stop any sound that is currently playing
          and any recording currently happening.
        - Call `stop()` or `write()` to stop the recording.
        """

        if not self.read_flag:
            raise SoundDeviceError(f"{self.device_name} does not support reading!")

        frames = int(time * self.sample_rate)
        recording = sd.rec(
            frames=frames,
            samplerate=self.sample_rate,
            channels=self.in_channels,
            device=self.device_number,
            blocking=blocking,
            dtype=self.config.dtype,
        )

        return AudioSegment(
            data=recording.flatten(),
            sample_width=recording.dtype.itemsize,
            frame_rate=self.sample_rate,
            channels=self.in_channels,
        )

    def stop(self):
        """
        Stop the sound device from playing or recording.

        Notes
        -----
        - This is a convenience function to stop the sound device from playing or recording.
        - It will stop any sound that is currently playing and any recording currently happening.
        """
        sd.stop()

    def wait(self):
        """
        Wait for the sound device to finish playing or recording.

        Notes
        -----
        - This is a convenience function to wait for the sound device to finish playing or recording.
        - It will block until the sound is played or recorded.
        """
        sd.wait()

    def open_write_stream(
        self,
        feed_data: Callable[[NDArray, int, Any, Any], None],
        on_done: Callable = lambda _: None,
        sample_rate: Optional[int] = None,
        channels: Optional[int] = None,
    ):
        if not self.write_flag or not self.stream_flag:
            raise SoundDeviceError(
                f"{self.device_name} does not support streaming writing!"
            )

        from sounddevice import CallbackFlags

        def callback(indata: NDArray, frames: int, time: Any, status: CallbackFlags):
            _ = frames
            flag_dict = {
                "input_overflow": status.input_overflow,
                "input_underflow": status.input_underflow,
                "output_overflow": status.output_overflow,
                "output_underflow": status.output_underflow,
                "priming_output": status.priming_output,
            }
            feed_data(indata, frames, time, flag_dict)

        try:
            assert sd is not None
            sample_rate = self.sample_rate if sample_rate is None else sample_rate
            channels = self.out_channels if channels is None else channels
            logging.warning(
                f"Opening Output Stream with sample_rate: {sample_rate} channels: {channels} device: {self.device_number} dtype: {self.config.dtype}"
            )
            self.out_stream = sd.OutputStream(
                samplerate=sample_rate,
                channels=channels,
                device=self.device_number,
                dtype=self.config.dtype,
                callback=callback,
                finished_callback=on_done,
            )
        except AttributeError:
            raise SoundDeviceError(
                f"Device {self.device_name} has not been correctly configured"
            )
        self.out_stream.start()

    def open_read_stream(
        self,
        on_feedback: Callable[[NDArray, dict[str, Any]], None],
        on_done: Callable = lambda _: None,
    ):
        if not self.read_flag or not self.stream_flag:
            raise SoundDeviceError(
                f"{self.device_name} does not support streaming reading!"
            )
        if self.in_stream is not None:
            raise SoundDeviceError(
                f"Stream for {self.device_name} is already open, close it first!"
            )
        from sounddevice import CallbackFlags

        def callback(indata: NDArray, frames: int, _, status: CallbackFlags):
            _ = frames
            indata = indata.flatten()
            sample_time_length = len(indata) / self.sample_rate
            if (
                self.sample_rate != self.config.consumer_sampling_rate
                and self.config.consumer_sampling_rate is not None
            ):
                indata = resample(
                    indata, int(sample_time_length * self.config.consumer_sampling_rate)
                )  # type: ignore
            flag_dict = {
                "input_overflow": status.input_overflow,
                "input_underflow": status.input_underflow,
                "output_overflow": status.output_overflow,
                "output_underflow": status.output_underflow,
                "priming_output": status.priming_output,
            }
            on_feedback(indata, flag_dict)

        try:
            if self.config.consumer_sampling_rate is None:
                window_size_samples = self.config.block_size * self.sample_rate
            else:
                window_size_samples = int(
                    self.config.block_size
                    * self.sample_rate
                    / self.config.consumer_sampling_rate
                )

            logging.warning(
                f"Opening Input Stream with sample_rate: {self.sample_rate} channels: {self.in_channels} device: {self.device_number} dtype: {self.config.dtype} blocksize: {window_size_samples}"
            )
            self.in_stream = sd.InputStream(
                samplerate=self.sample_rate,
                channels=self.in_channels,
                device=self.device_number,
                dtype=self.config.dtype,
                blocksize=window_size_samples,
                callback=callback,
                finished_callback=on_done,
            )
        except AttributeError:
            raise SoundDeviceError(
                f"Device {self.device_name} has not been correctly configured"
            )
        self.in_stream.start()

    def close_read_stream(self):
        if self.in_stream is not None:
            self.in_stream.stop()
            self.in_stream.close()
            self.in_stream = None

    def close_write_stream(self):
        if self.out_stream is not None:
            self.out_stream.stop()
            self.out_stream.close()
            self.out_stream = None



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import typing

import rosidl_adapter
import rosidl_adapter.parser
import rosidl_runtime_py.convert
from rosidl_adapter.parser import (
    ACTION_REQUEST_RESPONSE_SEPARATOR,
    SERVICE_REQUEST_RESPONSE_SEPARATOR,
    Constant,
    MessageSpecification,
    parse_message_string,
)


class InterfaceTextLine:
    """A convenience class for a single text line in an interface file."""

    def __init__(
        self,
        pkg_name: str,
        msg_name: str,
        line_text: str,
    ):
        if line_text in (
            SERVICE_REQUEST_RESPONSE_SEPARATOR,
            ACTION_REQUEST_RESPONSE_SEPARATOR,
        ):
            msg_spec = None
        else:
            msg_spec = parse_message_string(
                pkg_name=pkg_name,
                msg_name=msg_name,
                message_string=line_text,
            )
            if len(msg_spec.fields) > 1:  # type: ignore
                raise ValueError("'line_text' must be only one line")
        self._msg_spec: MessageSpecification | None = msg_spec
        self._raw_line_text = line_text

    def __str__(self) -> str:
        return self._raw_line_text

    def is_comment(self) -> bool:
        return bool(self._msg_spec) and self._msg_spec.annotations["comment"]  # type: ignore

    def is_trailing_comment(self) -> bool:
        return self._is_field_trailing_comment() or self._is_constant_trailing_comment()

    def _is_field_trailing_comment(self) -> bool:
        return self._field and self._field.annotations["comment"]  # type: ignore

    def _is_constant_trailing_comment(self) -> bool:
        return self._constant and self._constant.annotations["comment"]  # type: ignore

    @property
    def nested_type(self) -> typing.Optional[str]:
        if self._field and self._is_nested():
            interface_type: str = str(self._field.type)
            if self._field.type.is_array:  # type: ignore
                interface_type = interface_type[: interface_type.find("[")]
            return interface_type.replace("/", "/msg/")

    @property
    def trailing_comment(self) -> typing.Optional[str]:
        if self._is_field_trailing_comment():
            return self._field.annotations["comment"][0]  # type: ignore
        elif self._is_constant_trailing_comment():
            return self._constant.annotations["comment"][0]  # type: ignore
        else:
            return None

    @property
    def _field(self) -> rosidl_adapter.parser.Field | None:
        if self._msg_spec and self._msg_spec.fields:  # type: ignore
            return self._msg_spec.fields[0]  # type: ignore

    @property
    def _constant(self) -> Constant | None:
        if self._msg_spec and self._msg_spec.constants:  # type: ignore
            return self._msg_spec.constants[0]  # type: ignore

    def _is_nested(self) -> bool:
        if self._msg_spec and self._msg_spec.fields and self._field:  # type: ignore
            return "/" in str(self._field.type)
        else:
            return False


def _get_interface_lines(
    interface_identifier: str,
) -> typing.Iterable[InterfaceTextLine]:
    parts: typing.List[str] = interface_identifier.split("/")
    if len(parts) != 3:
        raise ValueError(
            f"Invalid name '{interface_identifier}'. Expected three parts separated by '/'"
        )
    pkg_name, _, msg_name = parts

    file_path = rosidl_runtime_py.get_interface_path(interface_identifier)
    with open(file_path) as file_handler:
        for line in file_handler:
            yield InterfaceTextLine(
                pkg_name=pkg_name,
                msg_name=msg_name,
                line_text=line.rstrip(),
            )


def _render_interface_line(
    line: InterfaceTextLine, is_show_comments: bool, indent_level: int
) -> str:
    text = str(line)
    if not is_show_comments:
        if not text or line.is_comment():
            return ""
        elif line.is_trailing_comment():
            if line.trailing_comment:
                comment_start_idx = text.find(line.trailing_comment)
                text = text[: comment_start_idx - 1].strip()
    if text:
        indent_string = indent_level * "\t"
        return f"{indent_string}{text}"
    return ""


def render_interface_string(
    interface_identifier: str,
    is_show_comments: bool = True,
    is_show_nested_comments: bool = False,
    indent_level: int = 0,
) -> str:
    lines: typing.List[str] = []
    for line in _get_interface_lines(interface_identifier):
        rendered = _render_interface_line(
            line, is_show_comments=is_show_comments, indent_level=indent_level
        )
        if rendered.strip():
            lines.append(rendered)
        if line.nested_type:
            nested_rendered = render_interface_string(
                line.nested_type,
                is_show_comments=is_show_nested_comments,
                is_show_nested_comments=is_show_nested_comments,
                indent_level=indent_level + 1,
            )
            if nested_rendered.strip():
                lines.append(nested_rendered)

    return "\n".join(lines)



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List

from langchain_core.messages import BaseMessage, HumanMessage

from rai.agents.conversational_agent import create_conversational_agent
from rai.initialization import get_llm_model
from rai.tools.ros2.cli import ros2_action, ros2_interface, ros2_service, ros2_topic


class Agent:
    def __init__(self):
        self.history: List[BaseMessage] = []
        self.llm = get_llm_model(model_type="complex_model")


class ROS2Agent(Agent):
    def __init__(self):
        super().__init__()
        self.tools = [
            ros2_topic,
            ros2_interface,
            ros2_service,
            ros2_action,
        ]
        self.agent = create_conversational_agent(
            self.llm, self.tools, "You are a ROS2 expert."
        )

    def __call__(self, message: str):
        self.history.append(HumanMessage(content=message))
        response = self.agent.invoke({"messages": self.history})
        output = response["messages"][-1].content
        return output



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rai.agents.base import BaseAgent
from rai.agents.conversational_agent import create_conversational_agent
from rai.agents.langchain.react_agent import ReActAgent
from rai.agents.runner import AgentRunner, wait_for_shutdown
from rai.agents.tool_runner import ToolRunner

__all__ = [
    "AgentRunner",
    "BaseAgent",
    "ReActAgent",
    "ToolRunner",
    "create_conversational_agent",
    "wait_for_shutdown",
]



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List

from rai.types.base import ROS2BaseModel
from rai.types.std import Header
from rai.types.vision import RegionOfInterest


class BaseSensorModel(ROS2BaseModel):
    _prefix: str = "sensor_msgs/msg"
    header: Header = Header()


class CameraInfo(BaseSensorModel):
    height: int = 0
    width: int = 0
    distortion_model: str = ""
    d: List[float] = []
    k: List[float] = [0.0] * 9
    r: List[float] = [0.0] * 9
    p: List[float] = [0.0] * 12
    binning_x: int = 0
    binning_y: int = 0
    roi: RegionOfInterest = RegionOfInterest()


class Image(BaseSensorModel):
    height: int = 0
    width: int = 0
    encoding: str = ""
    is_bigendian: int = 0
    step: int = 0
    data: List[int] = []



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Literal, TypedDict


class OpenAIKwargs(TypedDict):
    """Dictionary type to specify kwargs for OpenAI models."""

    model: str


class AWSBedrockKwargs(TypedDict):
    """Dictionary type to specify kwargs for AWS Bedrock models."""

    model_id: str
    region_name: Literal["us-west-1", "us-west-2"]


# OpenAI Models
OPENAI_MULTIMODAL: OpenAIKwargs = {
    "model": "gpt-4o-2024-08-06",
}

OPENAI_MINI: OpenAIKwargs = {
    "model": "gpt-4o-mini",
}

# AWS Bedrock Models
BEDROCK_CLAUDE_HAIKU: AWSBedrockKwargs = {
    "model_id": "anthropic.claude-3-haiku-20240307-v1:0",
    "region_name": "us-west-2",
}

BEDROCK_CLAUDE_SONNET: AWSBedrockKwargs = {
    "model_id": "anthropic.claude-3-5-sonnet-20240620-v1:0",
    "region_name": "us-east-1",
}

BEDROCK_CLAUDE_OPUS: AWSBedrockKwargs = {
    "model_id": "anthropic.claude-3-opus-20240229-v1:0",
    "region_name": "us-west-2",
}

BEDROCK_MULTIMODAL: AWSBedrockKwargs = {
    "model_id": "anthropic.claude-3-5-sonnet-20240620-v1:0",
    "region_name": "us-east-1",
}



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from langchain_core.language_models import BaseChatModel
from langchain_core.tools import BaseTool

from rai.agents.langchain.agent import LangChainAgent
from rai.agents.langchain.runnables import ReActAgentState, create_react_runnable
from rai.communication.hri_connector import HRIConnector, HRIMessage
from rai.messages.multimodal import SystemMultimodalMessage


class ReActAgent(LangChainAgent):
    def __init__(
        self,
        target_connectors: Dict[str, HRIConnector[HRIMessage]],
        llm: Optional[BaseChatModel] = None,
        tools: Optional[List[BaseTool]] = None,
        state: Optional[ReActAgentState] = None,
        system_prompt: Optional[str | SystemMultimodalMessage] = None,
    ):
        runnable = create_react_runnable(
            llm=llm, tools=tools, system_prompt=system_prompt
        )
        super().__init__(
            target_connectors=target_connectors,
            runnable=runnable,
            state=state,
        )



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Callable, Dict, List, Optional, Tuple

import rclpy.callback_groups
import rclpy.node
from rclpy.action.graph import get_action_names_and_types


class ROS2Context:
    """A context manager and decorator for ROS 2 node lifecycle management.

    This class ensures proper initialization and cleanup of the ROS 2 system,
    whether used as a context manager or decorator.

    Parameters
    ----------
    domain_id : int, optional
        ROS domain ID for network segmentation. Default is None.

    Attributes
    ----------
    domain_id : int or None
        The ROS domain ID used for initialization
    _initialized : bool
        Internal flag tracking ROS 2 initialization state

    Methods
    -------
    is_initialized
        Property that returns the current ROS 2 initialization state

    Examples
    --------
    Using as a context manager:

    >>> with ROS2Context() as ctx:
    ...     node = rclpy.create_node('my_node')
    ...     rclpy.spin_once(node)

    Using as a decorator:

    >>> @ROS2Context()
    ... def run_ros_node():
    ...     node = rclpy.create_node('my_node')
    ...     rclpy.spin_once(node)
    """

    def __init__(self, *, domain_id: Optional[int] = None):
        """Initialize the ROS2Context.

        Parameters
        ----------
        domain_id : int, optional
            ROS domain ID for network segmentation
        """
        self.domain_id = domain_id
        self._initialized = False

    def __enter__(self):
        """Enter the context manager.

        Returns
        -------
        ROS2Context
            The initialized context manager instance

        Raises
        ------
        RuntimeError
            If ROS 2 initialization fails
        """
        try:
            rclpy.init(domain_id=self.domain_id)
            self._initialized = True
            return self
        except Exception as e:
            raise RuntimeError(f"Failed to initialize ROS 2: {e}") from e

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager and cleanup ROS 2.

        Parameters
        ----------
        exc_type : type
            The type of the exception that occurred, if any
        exc_value : Exception
            The exception instance that occurred, if any
        traceback : traceback
            The traceback of the exception that occurred, if any
        """
        if self._initialized:
            try:
                rclpy.shutdown()
            except Exception as e:
                import logging

                logging.error(f"Error during ROS 2 shutdown: {e}")
        self._initialized = False

    def __call__(self, func):
        """Make the class callable as a decorator.

        Parameters
        ----------
        func : callable
            The function to be wrapped

        Returns
        -------
        callable
            The wrapped function with ROS 2 context management
        """

        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        # Preserve the original function's metadata
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper

    @property
    def is_initialized(self) -> bool:
        """Check if ROS 2 is currently initialized.

        Returns
        -------
        bool
            True if ROS 2 is initialized, False otherwise
        """
        return self._initialized


class NodeDiscovery:
    def __init__(
        self,
        node: rclpy.node.Node,
        allowlist: Optional[List[str]] = None,
        period_sec: float = 0.5,
        setters: Optional[List[Callable]] = None,
    ) -> None:
        self.period_sec = period_sec
        self.node = node

        self.topics_and_types: Dict[str, str] = dict()
        self.services_and_types: Dict[str, str] = dict()
        self.actions_and_types: Dict[str, str] = dict()
        self.allowlist: Optional[List[str]] = allowlist

        self.timer = self.node.create_timer(
            self.period_sec,
            self.discovery_callback,
            callback_group=rclpy.callback_groups.MutuallyExclusiveCallbackGroup(),
        )

        # callables (e.g. fun(x: NodeDiscovery)) that will receive the discovery
        # info on every timer callback. This allows to register other entities that
        # needs up-to-date discovery info
        if setters is None:
            self.setters = list()
        else:
            self.setters = setters

        # make first callback as fast as possible
        # sleep before first callback due ros discovery issue: https://github.com/ros2/ros2/issues/1057
        time.sleep(0.5)
        self.discovery_callback()

    def add_setter(self, setter: Callable):
        self.setters.append(setter)

    def discovery_callback(self):
        self.__set(
            self.node.get_topic_names_and_types(),
            self.node.get_service_names_and_types(),
            get_action_names_and_types(self.node),
        )
        for setter in self.setters:
            setter(self)

    def __set(self, topics, services, actions):
        def to_dict(info: List[Tuple[str, List[str]]]) -> Dict[str, str]:
            return {k: v[0] for k, v in info}

        self.topics_and_types = to_dict(topics)
        self.services_and_types = to_dict(services)
        self.actions_and_types = to_dict(actions)
        if self.allowlist is not None:
            self.__filter(self.allowlist)

    def __filter(self, allowlist: List[str]):
        for d in [
            self.topics_and_types,
            self.services_and_types,
            self.actions_and_types,
        ]:
            to_remove = [k for k in d if k not in allowlist]
            for k in to_remove:
                d.pop(k)

    def dict(self):
        return {
            "topics_and_types": self.topics_and_types,
            "services_and_types": self.services_and_types,
            "actions_and_types": self.actions_and_types,
        }



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import signal
from threading import Event
from typing import List

from rai.agents.base import BaseAgent

logger = logging.getLogger(__name__)


def wait_for_shutdown(agents: List[BaseAgent]):
    """Block until a shutdown signal (SIGINT or SIGTERM) is received, ensuring graceful termination.

    Parameters
    ----------
    agents : List[BaseAgent]
        List of running agents to be stopped on shutdown.

    Notes
    -----
    This method ensures a graceful shutdown of both the agent and the ROS2 node upon receiving
    an interrupt signal (SIGINT, e.g., Ctrl+C) or SIGTERM. It installs signal handlers to
    capture these events and invokes the agent's ``stop()`` method as part of the shutdown process.
    """
    shutdown_event = Event()

    def signal_handler(signum, frame):
        shutdown_event.set()

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    try:
        shutdown_event.wait()
    finally:
        for agent in agents:
            agent.stop()


def run_agents(agents: List[BaseAgent]):
    """Runs the agents in the background.

    Args:
        agents: List of agent instances
    """
    logger.info(
        "run_agents is an experimental function. \
                   If you believe that your agents are not running properly, \
                   please run them separately (in different processes)."
    )
    for agent in agents:
        agent.run()


class AgentRunner:
    """Runs the agents in the background.

    Parameters
    ----------
    agents : List[BaseAgent]
        List of agent instances
    """

    def __init__(self, agents: List[BaseAgent]):
        self.agents = agents
        self.logger = logging.getLogger(__name__)

    def run(self):
        self.logger.info(
            f"{self.__class__.__name__}.{self.run.__name__} is an experimental function. \
                            If you believe that your agents are not running properly, \
                            please run them separately (in different processes)."
        )
        for agent in self.agents:
            agent.run()

    def run_and_wait_for_shutdown(self):
        self.run()
        self.wait_for_shutdown()

    def wait_for_shutdown(self):
        shutdown_event = Event()

        def signal_handler(signum, frame):
            shutdown_event.set()

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        try:
            shutdown_event.wait()
        finally:
            for agent in self.agents:
                agent.stop()

    def stop(self):
        for agent in self.agents:
            agent.stop()



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import partial
from typing import Dict, List

import numpy as np
import requests
import sounddevice as sd
import streamlit as st
import tomli
import tomli_w
from elevenlabs import ElevenLabs
from langchain_aws import BedrockEmbeddings, ChatBedrock
from langchain_ollama import ChatOllama, OllamaEmbeddings
from langchain_openai import ChatOpenAI, OpenAIEmbeddings


def get_sound_devices(
    reinitialize: bool = False, output: bool = False
) -> List[Dict[str, str | int]]:
    if reinitialize:
        sd._terminate()
        sd._initialize()
    devices: List[Dict[str, str | int]] = sd.query_devices()
    if output:
        recording_devices = [
            device for device in devices if device.get("max_output_channels", 0) > 0
        ]
    else:
        recording_devices = [
            device for device in devices if device.get("max_input_channels", 0) > 0
        ]
    return recording_devices


def welcome():
    st.title("Welcome to RAI Configurator! 👋")
    st.markdown(
        """
    This wizard will help you set up your RAI environment step by step:
    1. Configure your AI models and vendor
    2. Set up model tracing and monitoring
    3. Configure speech recognition (ASR) (if installed)
    4. Set up text-to-speech (TTS) (if installed)
    5. Enable additional features
    6. Review and save your configuration

    Let's get started!
    """
    )

    st.button("Begin Configuration →", on_click=next_step)


def model_selection():
    st.title("Model Configuration")
    st.info(
        """
    This step configures which AI models will be used by RAI's agents. Different models have different capabilities and costs:
    - Simple models are faster and cheaper, used for basic tasks
    - Complex models are more capable but slower, used for complex reasoning
    - Embedding models convert text into numerical representations for memory and search
    """
    )

    def on_vendor_change():
        st.session_state.vendor_index = ["openai", "aws", "ollama"].index(
            st.session_state.vendor
        )

    vendor = st.selectbox(
        "Which AI vendor would you like to use?",
        ["openai", "aws", "ollama"],
        placeholder="Select vendor",
        key="vendor",
        index=st.session_state.get("vendor_index", 0),
        on_change=on_vendor_change,
    )

    if vendor:
        if vendor == "openai":
            st.write(
                f"Check out available {vendor} models [here](https://platform.openai.com/docs/models)"
            )
            simple_model = st.text_input(
                "Model for simple tasks",
                value=st.session_state["config"]["openai"]["simple_model"],
                key="simple_model",
            )
            complex_model = st.text_input(
                "Model for complex tasks",
                value=st.session_state["config"]["openai"]["complex_model"],
                key="complex_model",
            )
            embeddings_model = st.text_input(
                "Embeddings model",
                value=st.session_state["config"]["openai"]["embeddings_model"],
                key="embeddings_model",
            )

            def on_openai_compatible_api_change():
                st.session_state.use_openai_compatible_api = (
                    st.session_state.openai_compatible_api_checkbox
                )

            if "use_openai_compatible_api" not in st.session_state:
                st.session_state.use_openai_compatible_api = False

            use_openai_compatible_api = st.checkbox(
                "Use OpenAI compatible API",
                value=st.session_state.use_openai_compatible_api,
                key="openai_compatible_api_checkbox",
                on_change=on_openai_compatible_api_change,
            )
            st.session_state.use_openai_compatible_api = use_openai_compatible_api

            if use_openai_compatible_api:
                st.info(
                    "Used for OpenAI compatible endpoints, e.g. Ollama, vLLM... Make sure to specify `OPENAI_API_KEY` environment variable based on vendor's specification."
                )
                openai_api_base_url = st.text_input(
                    "OpenAI API base URL",
                    value=st.session_state["config"]["openai"]["base_url"],
                    key="openai_api_base_url",
                )
            else:
                openai_api_base_url = st.session_state["config"]["openai"]["base_url"]
            st.session_state.config["openai"] = {
                "simple_model": simple_model,
                "complex_model": complex_model,
                "embeddings_model": embeddings_model,
                "base_url": openai_api_base_url,
            }

        elif vendor == "aws":
            st.write(
                f"Check out available {vendor} models [here](https://docs.aws.amazon.com/bedrock/latest/userguide/model-ids.html)"
            )
            col1, col2 = st.columns(2)

            with col1:
                simple_model = st.text_input(
                    "Model for simple tasks",
                    value=st.session_state["config"]["aws"]["simple_model"],
                )
                complex_model = st.text_input(
                    "Model for complex tasks",
                    value=st.session_state["config"]["aws"]["complex_model"],
                )
                embeddings_model = st.text_input(
                    "Embeddings model",
                    value=st.session_state["config"]["aws"]["embeddings_model"],
                )
                st.session_state.config["aws"]["simple_model"] = simple_model
                st.session_state.config["aws"]["complex_model"] = complex_model
                st.session_state.config["aws"]["embeddings_model"] = embeddings_model

            with col2:
                aws_bedrock_region = st.text_input(
                    "AWS Bedrock region",
                    value=st.session_state["config"]["aws"]["region_name"],
                )
                st.session_state.config["aws"]["region_name"] = aws_bedrock_region

        elif vendor == "ollama":
            st.write(
                f"Check out available {vendor} models [here](https://ollama.com/models)"
            )
            base_url = st.text_input(
                "Ollama base URL",
                value=st.session_state["config"]["ollama"]["base_url"],
            )
            simple_model = st.text_input(
                "Model for simple tasks",
                value=st.session_state["config"]["ollama"]["simple_model"],
            )
            complex_model = st.text_input(
                "Model for complex tasks",
                value=st.session_state["config"]["ollama"]["complex_model"],
            )
            embeddings_model = st.text_input(
                "Embeddings model",
                value=st.session_state["config"]["ollama"]["embeddings_model"],
            )
            st.session_state.config["ollama"] = {
                "base_url": base_url,
                "simple_model": simple_model,
                "complex_model": complex_model,
                "embeddings_model": embeddings_model,
            }

    st.subheader("Multivendor configuration (Advanced)")
    st.write(
        "If you have access to multiple vendors, you can configure the models to use different vendors."
    )

    def on_advanced_config_change():
        st.session_state.use_advanced_config = st.session_state.advanced_config_checkbox

    if "use_advanced_config" not in st.session_state:
        st.session_state.use_advanced_config = False

    use_advanced_config = st.checkbox(
        "Use advanced configuration",
        value=st.session_state.use_advanced_config,
        key="advanced_config_checkbox",
        on_change=on_advanced_config_change,
    )
    st.session_state.use_advanced_config = use_advanced_config

    advanced_config = st.container()
    if use_advanced_config:
        with advanced_config:

            def on_model_vendor_change(model_type: str):
                st.session_state.config["vendor"][f"{model_type}_model"] = (
                    st.session_state[f"{model_type}_vendor_select"]
                )

            simple_model_vendor = st.selectbox(
                "Simple model vendor",
                options=["openai", "aws", "ollama"],
                index=["openai", "aws", "ollama"].index(
                    st.session_state.config["vendor"]["simple_model"]
                ),
                key="simple_vendor_select",
                on_change=lambda: on_model_vendor_change("simple"),
            )

            complex_model_vendor = st.selectbox(
                "Complex model vendor",
                options=["openai", "aws", "ollama"],
                index=["openai", "aws", "ollama"].index(
                    st.session_state.config["vendor"]["complex_model"]
                ),
                key="complex_vendor_select",
                on_change=lambda: on_model_vendor_change("complex"),
            )

            embeddings_model_vendor = st.selectbox(
                "Embeddings model vendor",
                options=["openai", "aws", "ollama"],
                index=["openai", "aws", "ollama"].index(
                    st.session_state.config["vendor"]["embeddings_model"]
                ),
                key="embeddings_vendor_select",
                on_change=lambda: on_model_vendor_change("embeddings"),
            )
    else:
        st.session_state.config["vendor"] = {
            "simple_model": vendor,
            "complex_model": vendor,
            "embeddings_model": vendor,
        }

    # Navigation buttons
    col1, col2 = st.columns([1, 1])
    with col1:
        st.button("← Back", on_click=prev_step)
    with col2:
        st.button("Next →", on_click=next_step)


def tracing():
    st.title("Tracing Configuration")
    st.info(
        """
    Tracing helps you monitor, debug, and analyze your AI assistant's conversations and performance.
    Both frameworks offer similar capabilities, but have different pricing models:
    - Langfuse is open-source and free to use (recommended)
    - LangSmith is a paid service from LangChain

    You can enable either or both services based on your needs.
    """
    )

    def on_langfuse_change():
        st.session_state.config["tracing"]["langfuse"]["use_langfuse"] = (
            st.session_state.langfuse_checkbox
        )

    def on_langfuse_host_change():
        st.session_state.config["tracing"]["langfuse"]["host"] = (
            st.session_state.langfuse_host_input
        )

    def on_langsmith_change():
        st.session_state.config["tracing"]["langsmith"]["use_langsmith"] = (
            st.session_state.langsmith_checkbox
        )

    # Ensure tracing config exists
    if "tracing" not in st.session_state.config:
        st.session_state.config["tracing"] = {}

    # Langfuse configuration
    st.subheader("Langfuse Configuration")
    langfuse_enabled = st.checkbox(
        "Enable Langfuse",
        value=st.session_state.config["tracing"]["langfuse"]["use_langfuse"],
        key="langfuse_checkbox",
        on_change=on_langfuse_change,
    )

    if langfuse_enabled:
        st.info(
            """
        Please ensure you have the following environment variables set:
        - `LANGFUSE_SECRET_KEY="sk-lf-..."`
        - `LANGFUSE_PUBLIC_KEY="pk-lf-..."`

        Find setup instructions [here](https://langfuse.com/docs/deployment/local)
        """
        )

        langfuse_host = st.text_input(
            "Langfuse Host",
            value=st.session_state.config["tracing"]["langfuse"]["host"],
            key="langfuse_host_input",
            on_change=on_langfuse_host_change,
        )

    # Langsmith configuration
    st.subheader("LangSmith Configuration")
    langsmith_enabled = st.checkbox(
        "Enable LangSmith",
        value=st.session_state.config["tracing"]["langsmith"]["use_langsmith"],
        key="langsmith_checkbox",
        on_change=on_langsmith_change,
    )

    if langsmith_enabled:
        st.info(
            """
        Please ensure you have the following environment variable set:
        - `LANGCHAIN_API_KEY`

        Find setup instructions [here](https://docs.smith.langchain.com/)
        """
        )

        # Store in config
        if "tracing" not in st.session_state.config:
            st.session_state.config["tracing"] = {}
        st.session_state.config["tracing"]["langsmith"] = {
            "use_langsmith": langsmith_enabled
        }

    # Navigation buttons
    col1, col2 = st.columns([1, 1])
    with col1:
        st.button("← Back", on_click=prev_step)
    with col2:
        st.button("Next →", on_click=next_step)


def asr():
    from rai_asr import TRANSCRIBE_MODELS

    def on_recording_device_change():
        st.session_state.config["asr"]["recording_device_name"] = (
            st.session_state.recording_device_select
        )

    def on_asr_vendor_change():
        st.session_state.config["asr"]["transcription_model"] = (
            st.session_state.asr_vendor_select
        )

    def on_model_name_change():
        st.session_state.config["asr"]["transcription_model_name"] = (
            st.session_state.model_name_input
        )

    def on_language_change():
        st.session_state.config["asr"]["language"] = st.session_state.language_input

    def on_silence_grace_change():
        st.session_state.config["asr"]["silence_grace_period"] = (
            st.session_state.silence_grace_input
        )

    def on_vad_threshold_change():
        st.session_state.config["asr"]["vad_threshold"] = (
            st.session_state.vad_threshold_input
        )

    def on_wake_word_change():
        st.session_state.config["asr"]["use_wake_word"] = (
            st.session_state.wake_word_checkbox
        )

    def on_wake_word_model_change():
        st.session_state.config["asr"]["wake_word_model"] = (
            st.session_state.wake_word_model_input
        )

    def on_wake_word_model_name_change():
        st.session_state.config["asr"]["wake_word_model_name"] = (
            st.session_state.wake_word_model_name_input
        )

    def on_wake_word_threshold_change():
        st.session_state.config["asr"]["wake_word_threshold"] = (
            st.session_state.wake_word_threshold_input
        )

    # Ensure asr config exists
    if "asr" not in st.session_state.config:
        st.session_state.config["asr"] = {}

    st.title("Speech Recognition Configuration")
    st.info(
        """
    Speech recognition (ASR - Automatic Speech Recognition) converts spoken words into text. This allows your assistant to understand voice input:
    - Local ASR uses Whisper and runs on your computer (recommended with GPU)
    - Device selection determines which microphone is used for voice input
    """
    )

    recording_devices = get_sound_devices()
    currently_selected_device_name = st.session_state.config.get("asr", {}).get(
        "recording_device_name", ""
    )
    try:
        device_index = [device["name"] for device in recording_devices].index(
            currently_selected_device_name
        )
    except ValueError:
        device_index = None

    recording_device_name = st.selectbox(
        "Default recording device",
        [device["name"] for device in recording_devices],
        placeholder="Select device",
        index=device_index,
        key="recording_device_select",
        on_change=on_recording_device_change,
    )

    refresh_devices = st.button("Refresh devices")
    if refresh_devices:
        recording_devices = get_sound_devices(reinitialize=True)

    # Get the current vendor from config and convert to display name
    current_vendor = st.session_state.config.get("asr", {}).get(
        "transciption_model", TRANSCRIBE_MODELS[0]
    )

    asr_vendor = st.selectbox(
        "Choose your ASR vendor",
        TRANSCRIBE_MODELS,
        placeholder="Select vendor",
        index=TRANSCRIBE_MODELS.index(current_vendor),
        key="asr_vendor_select",
        on_change=on_asr_vendor_change,
    )

    if asr_vendor == "OpenAI":
        st.info(
            """
        OpenAI ASR uses the OpenAI API. Make sure to set `OPENAI_API_KEY` environment variable.
        """
        )
    else:
        st.info(
            f"""
        {asr_vendor} is recommended to use when Nvidia GPU is available.
        """
        )

    # Add ASR parameters
    st.subheader("ASR Parameters")

    model_name = st.text_input(
        "Model name",
        value=st.session_state.config.get("asr", {}).get("model_name", "tiny"),
        help="Particular model architecture of the provided type, e.g. 'tiny'",
        key="model_name_input",
        on_change=on_model_name_change,
    )

    language = st.text_input(
        "Language code",
        value=st.session_state.config.get("asr", {}).get("language", "en"),
        help="Language code for the ASR model (e.g., 'en' for English)",
        key="language_input",
        on_change=on_language_change,
    )

    silence_grace_period = st.number_input(
        "Silence grace period (seconds)",
        value=float(
            st.session_state.config.get("asr", {}).get("silence_grace_period", 1.0)
        ),
        min_value=0.1,
        help="Grace period in seconds after silence to stop recording",
        key="silence_grace_input",
        on_change=on_silence_grace_change,
    )

    vad_threshold = st.slider(
        "VAD threshold",
        min_value=0.0,
        max_value=1.0,
        value=float(st.session_state.config.get("asr", {}).get("vad_threshold", 0.5)),
        help="Threshold for voice activity detection",
        key="vad_threshold_input",
        on_change=on_vad_threshold_change,
    )

    use_wake_word = st.checkbox(
        "Use wake word detection",
        value=st.session_state.config.get("asr", {}).get("use_wake_word", False),
        key="wake_word_checkbox",
        on_change=on_wake_word_change,
    )

    if use_wake_word:
        wake_word_model = st.text_input(
            "Wake word model",
            value=st.session_state.config.get("asr", {}).get("wake_word_model", ""),
            help="Wake word model type to use",
            key="wake_word_model_input",
            on_change=on_wake_word_model_change,
        )

        wake_word_model = st.text_input(
            "Wake word model name",
            value=st.session_state.config.get("asr", {}).get(
                "wake_word_model_name", ""
            ),
            help="Specific wake word model to use",
            key="wake_word_model_name_input",
            on_change=on_wake_word_model_name_change,
        )

        wake_word_threshold = st.slider(
            "Wake word threshold",
            min_value=0.0,
            max_value=1.0,
            value=float(
                st.session_state.config.get("asr", {}).get("wake_word_threshold", 0.5)
            ),
            help="Threshold for wake word detection",
            key="wake_word_threshold_input",
            on_change=on_wake_word_threshold_change,
        )

    col1, col2 = st.columns([1, 1])
    with col1:
        st.button("← Back", on_click=prev_step)
    with col2:
        st.button("Next →", on_click=next_step)


def tts():
    from rai_tts import TTS_MODELS

    def on_tts_vendor_change():
        st.session_state.config["tts"]["vendor"] = st.session_state.tts_vendor_select

    def on_voice_change():
        st.session_state.config["tts"]["voice"] = st.session_state.tts_voice_input

    def on_sound_device_change():
        st.session_state.config["tts"]["speaker_device_name"] = (
            st.session_state.sound_device_select
        )

    # Ensure tts config exists
    if "tts" not in st.session_state.config:
        st.session_state.config["tts"] = {}

    st.title("Text to Speech Configuration")
    st.info(
        """
    Text to Speech (TTS) converts your assistant's text responses into spoken words:
    - ElevenLabs provides high-quality, natural-sounding voices (requires API key)
    - OpenTTS runs locally on your computer with no API costs (requires Docker)
    """
    )

    sound_devices = get_sound_devices(output=True)
    currently_selected_device_name = st.session_state.config.get("tts", {}).get(
        "speaker_device_name", ""
    )
    try:
        device_index = [device["name"] for device in sound_devices].index(
            currently_selected_device_name
        )
    except ValueError:
        device_index = None

    recording_device_name = st.selectbox(
        "Default speaker device",
        [device["name"] for device in sound_devices],
        placeholder="Select device",
        index=device_index,
        key="sound_device_select",
        on_change=on_sound_device_change,
    )

    refresh_devices = st.button("Refresh devices")
    if refresh_devices:
        recording_devices = get_sound_devices(reinitialize=True, output=True)

    # Get the current vendor from config and convert to display name
    current_vendor = st.session_state.config.get("tts", {}).get("vendor", TTS_MODELS[0])

    tts_vendor = st.selectbox(
        "Choose your TTS vendor",
        TTS_MODELS,
        placeholder="Select vendor",
        index=TTS_MODELS.index(current_vendor),
        key="tts_vendor_select",
        on_change=on_tts_vendor_change,
    )

    if tts_vendor == "ElevenLabs":
        st.info(
            """
        Please ensure you have the following environment variable set:
        ```sh
        export ELEVENLABS_API_KEY="..."
        ```

        To get your API key, follow the instructions [here](https://elevenlabs.io/docs/api-reference/getting-started)
        """
        )
    elif tts_vendor == "OpenTTS":
        st.info(
            """
        Please ensure you have the Docker container running:
        ```sh
        docker run -it -p 5500:5500 synesthesiam/opentts:en
        ```

        To learn more about OpenTTS, visit [here](https://github.com/synesthesiam/opentts)
        """
        )

    model_name = st.text_input(
        "Voice",
        value=st.session_state.config.get("asr", {}).get("voice", ""),
        help="Voice compatible with selected vendor. If left empty RAI will select a deafault value.",
        key="tts_voice_input",
        on_change=on_voice_change,
    )

    st.info(
        """
    Some speakers enter power-saving mode when inactive.
    Enabling this option will keep the speaker active, reducing audio playback latency.
    """
    )

    col1, col2 = st.columns([1, 1])
    with col1:
        st.button("← Back", on_click=prev_step)
    with col2:
        st.button("Next →", on_click=next_step)


def additional_features():
    st.title("Additional Features Configuration")
    st.info(
        """
    These optional features enhance your assistant's capabilities but require additional dependencies.
    Follow the installation instructions below for the features you want to use.
    """
    )

    # OpenSET Instructions
    st.subheader("OpenSET (Visual Understanding)")
    st.markdown(
        """
    OpenSET provides visual understanding through Grounding DINO and Grounded SAM models.

    To install OpenSET dependencies, run:
    ```bash
    poetry install --with openset
    ```

    This will install:
    - Grounding DINO for object detection
    - Grounded SAM for segmentation
    - Required CUDA dependencies
    """
    )

    # NOMAD Instructions
    st.subheader("NOMAD (Navigation)")
    st.markdown(
        """
    NOMAD enables navigation capabilities using transformer-based image processing.

    To install NOMAD dependencies, run:
    ```bash
    poetry install --with nomad
    ```

    This will install:
    - NOMAD navigation transformer
    - Required image processing libraries
    """
    )

    st.info(
        "⚠️ Note: These features require significant disk space and may need a GPU for optimal performance."
    )

    # Navigation buttons
    col1, col2 = st.columns([1, 1])
    with col1:
        st.button("← Back", on_click=prev_step)
    with col2:
        st.button("Next →", on_click=next_step)


def review_and_save():
    st.title("Review & Save Configuration")
    st.write(
        """
    This is the final step where you can:
    - Review all your configuration settings
    - Test the configuration to ensure everything works
    - Save the settings to a file that your assistant will use
    """
    )
    st.info(
        "The configuration contains default values for each setting, even if you didn't set them."
    )

    # Display current configuration
    st.subheader("Current Configuration")
    toml_string = tomli_w.dumps(st.session_state.config)
    st.code(toml_string, language="toml")

    if st.button("Test Configuration"):
        progress = st.progress(0.0)
        test_results = {}

        def create_chat_model(model_type: str):
            vendor_name = st.session_state.config["vendor"][f"{model_type}_model"]
            model_name = st.session_state.config[vendor_name][f"{model_type}_model"]

            if vendor_name == "openai":
                base_url = st.session_state.config["openai"]["base_url"]
                return ChatOpenAI(model=model_name, base_url=base_url)
            elif vendor_name == "aws":
                return ChatBedrock(model_id=model_name)
            elif vendor_name == "ollama":
                return ChatOllama(
                    model=model_name,
                    base_url=st.session_state.config["ollama"]["base_url"],
                )
            raise ValueError(f"Unknown vendor: {vendor_name}")

        def test_chat_model(model_type: str) -> bool:
            try:
                model = create_chat_model(model_type)
                answer = model.invoke("Say hello!")
                return answer.content is not None
            except Exception as e:
                st.error(f"{model_type.title()} model error: {e}")
                return False

        def test_simple_model() -> bool:
            return test_chat_model("simple")

        def test_complex_model() -> bool:
            return test_chat_model("complex")

        def test_embeddings_model():
            try:
                embeddings_model_vendor_name = st.session_state.config["vendor"][
                    "embeddings_model"
                ]
                if embeddings_model_vendor_name == "openai":
                    embeddings_model = OpenAIEmbeddings(
                        model=st.session_state.config["openai"]["embeddings_model"]
                    )
                elif embeddings_model_vendor_name == "aws":
                    embeddings_model = BedrockEmbeddings(
                        model_id=st.session_state.config["aws"]["embeddings_model"]
                    )
                elif embeddings_model_vendor_name == "ollama":
                    embeddings_model = OllamaEmbeddings(
                        model=st.session_state.config["ollama"]["embeddings_model"],
                        base_url=st.session_state.config["ollama"]["base_url"],
                    )
                embeddings_answer = embeddings_model.embed_query("Say hello!")
                return embeddings_answer is not None
            except Exception as e:
                st.error(f"Embeddings model error: {e}")
                return False

        def test_langfuse():
            use_langfuse = st.session_state.config["tracing"]["langfuse"][
                "use_langfuse"
            ]
            if not use_langfuse:
                return True
            return bool(os.getenv("LANGFUSE_SECRET_KEY")) and bool(
                os.getenv("LANGFUSE_PUBLIC_KEY")
            )

        def test_langsmith():
            use_langsmith = st.session_state.config["tracing"]["langsmith"][
                "use_langsmith"
            ]
            if not use_langsmith:
                return True
            return bool(os.getenv("LANGCHAIN_API_KEY"))

        def test_tts():
            vendor = st.session_state.config["tts"]["vendor"]
            if vendor == "elevenlabs":
                try:
                    client = ElevenLabs(api_key=os.getenv("ELEVENLABS_API_KEY"))
                    output = client.generate(text="Hello, world!")
                    output = list(output)
                    return True
                except Exception as e:
                    st.error(f"TTS error: {e}")
                return False
            elif vendor == "opentts":
                try:
                    params = {
                        "voice": "glow-speak:en-us_mary_ann",
                        "text": "Hello, world!",
                    }
                    response = requests.get(
                        "http://localhost:5500/api/tts", params=params
                    )
                    if response.status_code == 200:
                        return True
                except Exception as e:
                    st.error(f"TTS error: {e}")
                return False

        def test_recording_device(index: int, sample_rate: int):
            try:
                recording = sd.rec(
                    device=index,
                    frames=sample_rate,
                    samplerate=sample_rate,
                    channels=1,
                    dtype="int16",
                )
                sd.wait()
                if np.sum(np.abs(recording)) == 0:
                    return False
                return True
            except Exception as e:
                st.error(f"Recording device error: {e}")
                return False

        # TODO: Add ASR test
        # TODO: Move tests to a separate file in tests/

        # Run tests

        devices = sd.query_devices()
        device_index = [device["name"] for device in devices].index(
            st.session_state.config["asr"]["recording_device_name"]
        )
        sample_rate = int(devices[device_index]["default_samplerate"])
        tests = [
            (test_simple_model, "Simple Model"),
            (test_complex_model, "Complex Model"),
            (test_embeddings_model, "Embeddings Model"),
            (test_langfuse, "Langfuse"),
            (test_langsmith, "LangSmith"),
            (test_tts, "TTS"),
            (
                partial(test_recording_device, device_index, sample_rate),
                "Recording Device",
            ),
        ]
        progress.progress(0.0, "Running tests...")
        for i, (test, name) in enumerate(tests):
            test_results[name] = test()
            progress.progress((1 + i) / len(tests), f"Testing {name}...")

        # Display results in a table
        st.subheader("Test Results")

        # Create a two-column table using streamlit columns
        col1, col2 = st.columns(2)
        with col1:
            st.markdown("**Component**")
            for component in test_results.keys():
                st.write(component)
        with col2:
            st.markdown("**Status**")
            for result in test_results.values():
                st.write("✅ Pass" if result else "❌ Fail")

        # Overall success message
        if all(test_results.values()):
            st.success("All tests passed! You can save the configuration now.")
        else:
            st.error("Some tests failed. Please check the errors above.")

    col1, col2 = st.columns([1, 1])
    with col1:
        st.button("← Back", on_click=prev_step)
    with col2:
        if st.button("Save Configuration"):
            # Save configuration to TOML file
            with open("config.toml", "wb") as f:
                tomli_w.dump(st.session_state.config, f)
            st.success("Configuration saved successfully!")


@st.cache_data
def setup_steps():
    step_names = ["👋 Welcome", "🤖 Model Selection", "📊 Tracing"]
    step_render = [welcome, model_selection, tracing]

    try:
        from rai_asr import TRANSCRIBE_MODELS

        step_names.append("🎙️ Speech Recognition")
        step_render.append(asr)
    except ImportError:
        pass

    try:
        from rai_tts import TTS_MODELS

        step_names.append("🔊 Text to Speech")
        step_render.append(tts)
    except ImportError:
        pass

    step_names.extend(
        [
            "🎯 Additional Features",
            "✅ Review & Save",
        ]
    )
    step_render.extend([additional_features, review_and_save])

    steps = dict(enumerate(step_names))
    step_renderer = dict(enumerate(step_render))
    return steps, step_renderer


# Initialize session state for tracking steps if not exists
if "current_step" not in st.session_state:
    st.session_state.current_step = 1
if "config" not in st.session_state:
    # Load initial config from TOML file
    try:
        with open("config.toml", "rb") as f:
            st.session_state.config = tomli.load(f)
    except FileNotFoundError:
        raise FileNotFoundError("config.toml not found. Please recreate it.")

# Sidebar progress tracker
st.sidebar.title("Configuration Progress")
steps, step_renderer = setup_steps()

# Replace the existing step display with clickable elements
for step_num, step_name in steps.items():
    if step_num == st.session_state.current_step:
        # Current step is bold and has an arrow
        if st.sidebar.button(
            step_name, key=f"step_{step_num}", use_container_width=True
        ):
            st.session_state.current_step = step_num
    else:
        # Other steps are clickable but not highlighted
        if st.sidebar.button(
            step_name, key=f"step_{step_num}", use_container_width=True
        ):
            st.session_state.current_step = step_num


# Navigation buttons
def next_step():
    st.session_state.current_step = st.session_state.current_step + 1


def prev_step():
    st.session_state.current_step = st.session_state.current_step - 1


# Main content based on current step
step_renderer[st.session_state.current_step]()



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydantic import BaseModel, ConfigDict


class RaiBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class ROS2BaseModel(RaiBaseModel):
    """
    Base model for messages replicated from ros2

    Attributes:
    _prefix: str
        Prefix of the ros2 message, for example std_msgs/msg
    """

    _prefix: str

    def get_msg_name(self) -> str:
        return f"{self._prefix}/{self.__class__.__name__}"



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import (
    Any,
    Dict,
    List,
    Protocol,
    Tuple,
    Type,
    runtime_checkable,
)

import rclpy
import rclpy.action
import rclpy.callback_groups
import rclpy.executors
import rclpy.node
import rclpy.qos
import rclpy.subscription
import rclpy.task
import rosidl_runtime_py.convert
import rosidl_runtime_py.set_message
import rosidl_runtime_py.utilities
from rclpy.qos import (
    DurabilityPolicy,
    HistoryPolicy,
    LivelinessPolicy,
    QoSProfile,
    ReliabilityPolicy,
)
from rclpy.topic_endpoint_info import TopicEndpointInfo
from rosidl_parser.definition import NamespacedType
from rosidl_runtime_py.import_message import import_message_from_namespaced_type
from rosidl_runtime_py.utilities import get_namespaced_type

from rai.communication.ros2.api.conversion import import_message_from_str


@runtime_checkable
class IROS2Message(Protocol):
    __slots__: list

    def get_fields_and_field_types(self) -> dict: ...


class BaseROS2API:
    node: rclpy.node.Node

    @staticmethod
    def adapt_requests_to_offers(publisher_info: List[TopicEndpointInfo]) -> QoSProfile:
        if not publisher_info:
            return QoSProfile(depth=1)

        num_endpoints = len(publisher_info)
        reliability_reliable_count = 0
        durability_transient_local_count = 0

        for endpoint in publisher_info:
            profile = endpoint.qos_profile
            if profile.reliability == ReliabilityPolicy.RELIABLE:
                reliability_reliable_count += 1
            if profile.durability == DurabilityPolicy.TRANSIENT_LOCAL:
                durability_transient_local_count += 1

        request_qos = QoSProfile(
            history=HistoryPolicy.KEEP_LAST,
            depth=1,
            liveliness=LivelinessPolicy.AUTOMATIC,
        )

        # Set reliability based on publisher offers
        if reliability_reliable_count == num_endpoints:
            request_qos.reliability = ReliabilityPolicy.RELIABLE
        else:
            if reliability_reliable_count > 0:
                logging.warning(
                    "Some, but not all, publishers are offering RELIABLE reliability. "
                    "Falling back to BEST_EFFORT as it will connect to all publishers. "
                    "Some messages from Reliable publishers could be dropped."
                )
            request_qos.reliability = ReliabilityPolicy.BEST_EFFORT

        # Set durability based on publisher offers
        if durability_transient_local_count == num_endpoints:
            request_qos.durability = DurabilityPolicy.TRANSIENT_LOCAL
        else:
            if durability_transient_local_count > 0:
                logging.warning(
                    "Some, but not all, publishers are offering TRANSIENT_LOCAL durability. "
                    "Falling back to VOLATILE as it will connect to all publishers. "
                    "Previously-published latched messages will not be retrieved."
                )
            request_qos.durability = DurabilityPolicy.VOLATILE

        return request_qos

    @staticmethod
    def build_ros2_msg(
        msg_type: str | type[rclpy.node.MsgType], msg_args: Dict[str, Any]
    ) -> object:
        """Build a ROS2 message instance from string or MsgType and content dictionary."""
        if isinstance(msg_type, str):
            msg_cls = import_message_from_str(msg_type)
        else:
            msg_cls = msg_type
        msg = msg_cls()
        rosidl_runtime_py.set_message.set_message_fields(msg, msg_args)
        return msg

    @staticmethod
    def build_ros2_service_request(
        service_type: str, service_request_args: Dict[str, Any]
    ) -> Tuple[object, Type[Any]]:
        msg_cls = import_message_from_str(service_type)
        msg = msg_cls.Request()
        rosidl_runtime_py.set_message.set_message_fields(msg, service_request_args)
        return msg, msg_cls  # type: ignore

    @staticmethod
    def import_message_from_str(msg_type: str) -> Type[object]:
        msg_namespaced_type: NamespacedType = get_namespaced_type(msg_type)
        return import_message_from_namespaced_type(msg_namespaced_type)

    def get_topic_type(self, topic: str) -> str:
        names_and_types = self.node.get_topic_names_and_types(no_demangle=False)
        for name, types in names_and_types:
            if name == (topic if topic.startswith("/") else f"/{topic}"):
                if len(types) != 1:
                    raise ValueError(f"Topic {topic} has multiple types: {types}")
                return types[0]
        raise ValueError(f"Topic {topic} not found")



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import logging
import time
from typing import Any, Callable, Dict, List, Optional, Sequence, Union, cast

from langchain_core.messages import AIMessage, ToolCall, ToolMessage
from langchain_core.runnables import RunnableConfig
from langchain_core.runnables.config import get_executor_for_config
from langchain_core.tools import BaseTool
from langchain_core.tools import tool as create_tool
from langgraph.prebuilt.tool_node import msg_content_output
from langgraph.utils.runnable import RunnableCallable
from pydantic import ValidationError

from rai.messages import MultimodalArtifact, ToolMultimodalMessage, store_artifacts


class ToolRunner(RunnableCallable):
    def __init__(
        self,
        tools: Sequence[Union[BaseTool, Callable]],
        *,
        name: str = "tools",
        tags: Optional[list[str]] = None,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        super().__init__(self._func, name=name, tags=tags, trace=False)
        self.logger = logger or logging.getLogger(__name__)
        self.tools_by_name: Dict[str, BaseTool] = {}
        for tool_ in tools:
            if not isinstance(tool_, BaseTool):
                tool_ = create_tool(tool_)
            self.tools_by_name[tool_.name] = tool_

    def _func(self, input: dict[str, Any], config: RunnableConfig) -> Any:
        config["max_concurrency"] = (
            1  # TODO(maciejmajek): use better mechanism for task queueing
        )
        if messages := input.get("messages", []):
            message = messages[-1]
        else:
            raise ValueError("No message found in input")

        if not isinstance(message, AIMessage):
            raise ValueError("Last message is not an AIMessage")

        def run_one(call: ToolCall):
            self.logger.info(f"Running tool: {call['name']}, args: {call['args']}")
            artifact = None

            try:
                ts = time.perf_counter()
                output = self.tools_by_name[call["name"]].invoke(call, config)  # type: ignore
                te = time.perf_counter() - ts
                self.logger.info(
                    f"Tool {call['name']} completed in {te:.2f} seconds. Tool output: {str(output.content)[:100]}{'...' if len(str(output.content)) > 100 else ''}"
                )
                self.logger.debug(
                    f"Tool {call['name']} output: \n\n{str(output.content)}"
                )
            except ValidationError as e:
                errors = e.errors()
                for error in errors:
                    error.pop(
                        "url"
                    )  # get rid of the  https://errors.pydantic.dev/... url

                error_message = f"""
                                    Validation error in tool {call["name"]}:
                                    {e.title}
                                    Number of errors: {e.error_count()}
                                    Errors:
                                    {json.dumps(errors, indent=2)}
                                """
                self.logger.info(error_message)
                output = ToolMessage(
                    content=error_message,
                    name=call["name"],
                    tool_call_id=call["id"],
                    status="error",
                )
            except Exception as e:
                self.logger.info(f'Error in "{call["name"]}", error: {e}')
                output = ToolMessage(
                    content=f"Failed to run tool. Error: {e}",
                    name=call["name"],
                    tool_call_id=call["id"],
                    status="error",
                )

            if output.artifact is not None:
                artifact = output.artifact
                if not isinstance(artifact, dict):
                    raise ValueError(
                        "Artifact must be a dictionary with optional keys: 'images', 'audios'"
                    )

                artifact = cast(MultimodalArtifact, artifact)
                store_artifacts(output.tool_call_id, [artifact])

            if artifact is not None and (
                len(artifact.get("images", [])) > 0
                or len(artifact.get("audios", [])) > 0
            ):  # multimodal case, we currently support images and audios artifacts
                return ToolMultimodalMessage(
                    content=msg_content_output(output.content),
                    name=call["name"],
                    tool_call_id=call["id"],
                    images=artifact.get("images", []),
                    audios=artifact.get("audios", []),
                )

            return output

        with get_executor_for_config(config) as executor:
            raw_outputs = [*executor.map(run_one, message.tool_calls)]
            outputs: List[Any] = []
            for raw_output in raw_outputs:
                if isinstance(raw_output, ToolMultimodalMessage):
                    outputs.extend(
                        raw_output.postprocess()
                    )  # openai please allow tool messages with images!
                else:
                    outputs.append(raw_output)

            # because we can't answer an aiMessage with an alternating sequence of tool and human messages
            # we sort the messages by type so that the tool messages are sent first
            # for more information see implementation of ToolMultimodalMessage.postprocess
            outputs.sort(key=lambda x: x.__class__.__name__, reverse=True)
            input["messages"].extend(outputs)
            return input



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time
import uuid
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar

import rclpy
import rclpy.executors
import rclpy.node
import rclpy.time
from rclpy.duration import Duration
from rclpy.executors import MultiThreadedExecutor
from rclpy.node import Node
from rclpy.qos import QoSProfile
from tf2_ros import Buffer, LookupException, TransformListener, TransformStamped

from rai.communication import BaseConnector
from rai.communication.ros2.api import (
    ROS2ActionAPI,
    ROS2ServiceAPI,
    ROS2TopicAPI,
)
from rai.communication.ros2.connectors.action_mixin import ROS2ActionMixin
from rai.communication.ros2.connectors.service_mixin import ROS2ServiceMixin
from rai.communication.ros2.messages import ROS2Message

T = TypeVar("T", bound=ROS2Message)


class ROS2BaseConnector(ROS2ActionMixin, ROS2ServiceMixin, BaseConnector[T]):
    """ROS2-specific implementation of the BaseConnector.

    This connector provides functionality for ROS2 communication through topics,
    services, and actions, as well as TF (Transform) operations.

    Parameters
    ----------
    node_name : str, optional
        Name of the ROS2 node. If not provided, generates a unique name with UUID.
    destroy_subscribers : bool, optional
        Whether to destroy subscribers after receiving a message, by default False.

    Methods
    -------
    get_topics_names_and_types()
        Get list of available topics and their message types.
    get_services_names_and_types()
        Get list of available services and their types.
    get_actions_names_and_types()
        Get list of available actions and their types.
    send_message(message, target, msg_type, auto_qos_matching=True, qos_profile=None, **kwargs)
        Send a message to a specified topic.
    receive_message(source, timeout_sec=1.0, msg_type=None, auto_topic_type=True, **kwargs)
        Receive a message from a specified topic.
    wait_for_transform(tf_buffer, target_frame, source_frame, timeout_sec=1.0)
        Wait for a transform to become available.
    get_transform(target_frame, source_frame, timeout_sec=5.0)
        Get the transform between two frames.
    create_service(service_name, on_request, on_done=None, service_type, **kwargs)
        Create a ROS2 service.
    create_action(action_name, generate_feedback_callback, action_type, **kwargs)
        Create a ROS2 action server.
    shutdown()
        Clean up resources and shut down the connector.

    Notes
    -----
    Threading Model:
        The connector creates a MultiThreadedExecutor that runs in a dedicated thread.
        This executor processes all ROS2 callbacks and operations asynchronously.

    Subscriber Lifecycle:
        The `destroy_subscribers` parameter controls subscriber cleanup behavior:
        - True: Subscribers are destroyed after receiving a message
            - Pros: Better resource utilization
            - Cons: Known stability issues (see: https://github.com/ros2/rclpy/issues/1142)
        - False (default): Subscribers remain active after message reception
            - Pros: More stable operation, avoids potential crashes
            - Cons: May lead to memory/performance overhead from inactive subscribers
    """

    def __init__(
        self,
        node_name: str = f"rai_ros2_connector_{str(uuid.uuid4())[-12:]}",
        destroy_subscribers: bool = False,
    ):
        super().__init__()

        if not rclpy.ok():
            rclpy.init()
            self.logger.warning(
                "Auto-initializing ROS2, but manual initialization is recommended. "
                "For better control and predictability, call rclpy.init() or ROS2Context before creating this connector."
            )

        self._node = Node(node_name)
        self._topic_api = ROS2TopicAPI(self._node, destroy_subscribers)
        self._service_api = ROS2ServiceAPI(self._node)
        self._actions_api = ROS2ActionAPI(self._node)
        self._tf_buffer = Buffer(node=self._node)
        self._tf_listener = TransformListener(self._tf_buffer, self._node)

        self._executor = MultiThreadedExecutor()
        self._executor.add_node(self._node)
        self._thread = threading.Thread(target=self._executor.spin)
        self._thread.start()

        # cache for last received messages
        self.last_msg: Dict[str, T] = {}

    def last_message_callback(self, source: str, msg: T):
        self.last_msg[source] = msg

    def get_topics_names_and_types(self) -> List[Tuple[str, List[str]]]:
        return self._topic_api.get_topic_names_and_types()

    def get_services_names_and_types(self) -> List[Tuple[str, List[str]]]:
        return self._service_api.get_service_names_and_types()

    def get_actions_names_and_types(self) -> List[Tuple[str, List[str]]]:
        return self._actions_api.get_action_names_and_types()

    def send_message(
        self,
        message: T,
        target: str,
        *,
        msg_type: str,  # TODO: allow msg_type to be None, add auto topic type detection
        auto_qos_matching: bool = True,
        qos_profile: Optional[QoSProfile] = None,
        **kwargs: Any,
    ):
        self._topic_api.publish(
            topic=target,
            msg_content=message.payload,
            msg_type=msg_type,
            auto_qos_matching=auto_qos_matching,
            qos_profile=qos_profile,
        )

    def general_callback_preprocessor(self, message: Any) -> T:
        return self.T_class(payload=message, metadata={"msg_type": str(type(message))})

    def register_callback(
        self,
        source: str,
        callback: Callable[[T | Any], None],
        raw: bool = False,
        *,
        msg_type: Optional[str] = None,
        qos_profile: Optional[QoSProfile] = None,
        auto_qos_matching: bool = True,
        **kwargs: Any,
    ) -> str:
        exists = self._topic_api.subscriber_exists(source)
        if not exists:
            self._topic_api.create_subscriber(
                topic=source,
                msg_type=msg_type,
                callback=partial(self.general_callback, source),
                qos_profile=qos_profile,
                auto_qos_matching=auto_qos_matching,
            )
        return super().register_callback(source, callback, raw=raw)

    def receive_message(
        self,
        source: str,
        timeout_sec: float = 1.0,
        *,
        msg_type: Optional[str] = None,
        qos_profile: Optional[QoSProfile] = None,
        auto_qos_matching: bool = True,
        **kwargs: Any,
    ) -> T:
        if self._topic_api.subscriber_exists(source):
            # trying to hit cache first
            if source in self.last_msg:
                if self.last_msg[source].timestamp > time.time() - timeout_sec:
                    return self.last_msg[source]
        else:
            self._topic_api.create_subscriber(
                topic=source,
                callback=partial(self.general_callback, source),
                msg_type=msg_type,
                qos_profile=qos_profile,
                auto_qos_matching=auto_qos_matching,
            )
            self.register_callback(source, partial(self.last_message_callback, source))

        start_time = time.time()
        # wait for the message to be received
        while time.time() - start_time < timeout_sec:
            if source in self.last_msg:
                return self.last_msg[source]
            time.sleep(0.1)
        else:
            raise TimeoutError(
                f"Message from {source} not received in {timeout_sec} seconds"
            )

    @staticmethod
    def wait_for_transform(
        tf_buffer: Buffer,
        target_frame: str,
        source_frame: str,
        timeout_sec: float = 1.0,
    ) -> bool:
        start_time = time.time()
        while time.time() - start_time < timeout_sec:
            if tf_buffer.can_transform(target_frame, source_frame, rclpy.time.Time()):
                return True
            time.sleep(0.1)
        return False

    def get_transform(
        self,
        target_frame: str,
        source_frame: str,
        timeout_sec: float = 5.0,
    ) -> TransformStamped:
        transform_available = self.wait_for_transform(
            self._tf_buffer, target_frame, source_frame, timeout_sec
        )
        if not transform_available:
            raise LookupException(
                f"Could not find transform from {source_frame} to {target_frame} in {timeout_sec} seconds"
            )
        transform: TransformStamped = self._tf_buffer.lookup_transform(
            target_frame,
            source_frame,
            rclpy.time.Time(),
            timeout=Duration(seconds=int(timeout_sec)),
        )

        return transform

    def create_service(
        self,
        service_name: str,
        on_request: Callable[[Any, Any], Any],
        on_done: Optional[Callable[[Any, Any], Any]] = None,
        *,
        service_type: str,
        **kwargs: Any,
    ) -> str:
        return self._service_api.create_service(
            service_name=service_name,
            callback=on_request,
            service_type=service_type,
            **kwargs,
        )

    def create_action(
        self,
        action_name: str,
        generate_feedback_callback: Callable,
        *,
        action_type: str,
        **kwargs: Any,
    ) -> str:
        return self._actions_api.create_action_server(
            action_name=action_name,
            action_type=action_type,
            execute_callback=generate_feedback_callback,
            **kwargs,
        )

    @property
    def node(self) -> Node:
        return self._node

    def shutdown(self):
        self._tf_listener.unregister()
        self._node.destroy_node()
        self._actions_api.shutdown()
        self._topic_api.shutdown()
        self._executor.shutdown()
        self._thread.join()



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from langchain_core.messages import (
    AIMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)

from .artifacts import MultimodalArtifact, get_stored_artifacts, store_artifacts
from .conversion import preprocess_image
from .multimodal import (
    AIMultimodalMessage,
    HumanMultimodalMessage,
    MultimodalMessage,
    SystemMultimodalMessage,
    ToolMultimodalMessage,
)

__all__ = [
    "AIMessage",
    "AIMultimodalMessage",
    "HumanMessage",
    "HumanMultimodalMessage",
    "MultimodalArtifact",
    "MultimodalMessage",
    "SystemMessage",
    "SystemMultimodalMessage",
    "ToolMessage",
    "ToolMultimodalMessage",
    "get_stored_artifacts",
    "preprocess_image",
    "store_artifacts",
]



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from dataclasses import dataclass
from typing import Any, Dict, List, Literal, Optional, Tuple, cast

import coloredlogs
import tomli
from langchain_aws import ChatBedrock
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.embeddings import Embeddings
from langchain_core.tracers.langchain import LangChainTracer
from langchain_ollama import ChatOllama
from langchain_openai import ChatOpenAI
from langsmith import Client

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
coloredlogs.install(level="INFO")  # type: ignore


@dataclass
class VendorConfig:
    simple_model: str
    complex_model: str
    embeddings_model: str


@dataclass
class ModelConfig:
    simple_model: str
    complex_model: str
    embeddings_model: str


@dataclass
class AWSConfig(ModelConfig):
    region_name: str


@dataclass
class OllamaConfig(ModelConfig):
    base_url: str


@dataclass
class OpenAIConfig(ModelConfig):
    base_url: str


@dataclass
class LangfuseConfig:
    use_langfuse: bool
    host: str


@dataclass
class LangsmithConfig:
    use_langsmith: bool
    host: str


@dataclass
class TracingConfig:
    project: str
    langfuse: LangfuseConfig
    langsmith: LangsmithConfig


@dataclass
class RAIConfig:
    vendor: VendorConfig
    aws: AWSConfig
    openai: OpenAIConfig
    ollama: OllamaConfig
    tracing: TracingConfig


def load_config(config_path: Optional[str] = None) -> RAIConfig:
    if config_path is None:
        with open("config.toml", "rb") as f:
            config_dict = tomli.load(f)
    else:
        with open(config_path, "rb") as f:
            config_dict = tomli.load(f)
    return RAIConfig(
        vendor=VendorConfig(**config_dict["vendor"]),
        aws=AWSConfig(**config_dict["aws"]),
        openai=OpenAIConfig(**config_dict["openai"]),
        ollama=OllamaConfig(**config_dict["ollama"]),
        tracing=TracingConfig(
            project=config_dict["tracing"]["project"],
            langfuse=LangfuseConfig(**config_dict["tracing"]["langfuse"]),
            langsmith=LangsmithConfig(**config_dict["tracing"]["langsmith"]),
        ),
    )


def get_llm_model_config_and_vendor(
    model_type: Literal["simple_model", "complex_model"],
    vendor: Optional[str] = None,
    config_path: Optional[str] = None,
) -> Tuple[str, str]:
    config = load_config(config_path)
    if vendor is None:
        if model_type == "simple_model":
            vendor = config.vendor.simple_model
        else:
            vendor = config.vendor.complex_model

    model_config = getattr(config, vendor)
    return model_config, vendor


def get_llm_model(
    model_type: Literal["simple_model", "complex_model"],
    vendor: Optional[str] = None,
    config_path: Optional[str] = None,
    **kwargs: Any,
) -> ChatOpenAI | ChatBedrock | ChatOllama:
    model_config, vendor = get_llm_model_config_and_vendor(
        model_type, vendor, config_path
    )
    model = getattr(model_config, model_type)
    logger.info(f"Initializing {model_type}: Vendor: {vendor}, Model: {model}")
    if vendor == "openai":
        from langchain_openai import ChatOpenAI

        model_config = cast(OpenAIConfig, model_config)

        return ChatOpenAI(model=model, base_url=model_config.base_url, **kwargs)
    elif vendor == "aws":
        from langchain_aws import ChatBedrock

        model_config = cast(AWSConfig, model_config)

        return ChatBedrock(
            model_id=model,
            region_name=model_config.region_name,
            **kwargs,
        )
    elif vendor == "ollama":
        from langchain_ollama import ChatOllama

        model_config = cast(OllamaConfig, model_config)
        return ChatOllama(model=model, base_url=model_config.base_url, **kwargs)
    else:
        raise ValueError(f"Unknown LLM vendor: {vendor}")


def get_llm_model_direct(
    model_name: str,
    vendor: str,
    config_path: Optional[str] = None,
    **kwargs: Any,
) -> ChatOpenAI | ChatBedrock | ChatOllama:
    config = load_config(config_path)
    model_config = getattr(config, vendor)

    logger.info(f"Initializing Model: {model_name}, Vendor: {vendor}")
    if vendor == "openai":
        from langchain_openai import ChatOpenAI

        model_config = cast(OpenAIConfig, model_config)

        return ChatOpenAI(model=model_name, base_url=model_config.base_url, **kwargs)
    elif vendor == "aws":
        from langchain_aws import ChatBedrock

        model_config = cast(AWSConfig, model_config)

        return ChatBedrock(
            model_id=model_name,
            region_name=model_config.region_name,
            **kwargs,
        )
    elif vendor == "ollama":
        from langchain_ollama import ChatOllama

        model_config = cast(OllamaConfig, model_config)
        return ChatOllama(model=model_name, base_url=model_config.base_url, **kwargs)
    else:
        raise ValueError(f"Unknown LLM vendor: {vendor}")


def get_embeddings_model(
    config_path: Optional[str] = None,
    return_kwargs: bool = False,
) -> Embeddings | Tuple[Embeddings, Dict[str, Any]]:
    config = load_config(config_path)
    vendor = config.vendor.embeddings_model

    model_config = getattr(config, vendor)

    logger.info(f"Using embeddings model: {vendor}-{model_config.embeddings_model}")
    if vendor == "openai":
        from langchain_openai import OpenAIEmbeddings

        model_config = cast(OpenAIConfig, model_config)
        embeddings = OpenAIEmbeddings(model=model_config.embeddings_model)
        if return_kwargs:
            c = (
                str(embeddings.__class__)
                .strip("<>")
                .replace("class '", "")
                .replace("'", "")
            )
            return embeddings, {
                "class": c,
                "model": model_config.embeddings_model,
                "vendor": vendor,
            }
        return embeddings

    elif vendor == "aws":
        from langchain_aws import BedrockEmbeddings

        model_config = cast(AWSConfig, model_config)
        embeddings = BedrockEmbeddings(
            model_id=model_config.embeddings_model, region_name=model_config.region_name
        )
        if return_kwargs:
            c = (
                str(embeddings.__class__)
                .strip("<>")
                .replace("class '", "")
                .replace("'", "")
            )
            return embeddings, {
                "class": c,
                "model_id": model_config.embeddings_model,
                "region_name": model_config.region_name,
                "vendor": vendor,
            }
        return embeddings

    elif vendor == "ollama":
        from langchain_ollama import OllamaEmbeddings

        model_config = cast(OllamaConfig, model_config)
        embeddings = OllamaEmbeddings(
            model=model_config.embeddings_model, base_url=model_config.base_url
        )
        if return_kwargs:
            c = (
                str(embeddings.__class__)
                .strip("<>")
                .replace("class '", "")
                .replace("'", "")
            )
            return embeddings, {
                "class": c,
                "model": model_config.embeddings_model,
                "base_url": model_config.base_url,
                "vendor": vendor,
            }
        return embeddings
    else:
        raise ValueError(f"Unknown embeddings vendor: {vendor}")


def get_tracing_callbacks(
    override_use_langfuse: bool = False, override_use_langsmith: bool = False
) -> List[BaseCallbackHandler]:
    config = load_config()
    callbacks: List[BaseCallbackHandler] = []
    if config.tracing.langfuse.use_langfuse or override_use_langfuse:
        from langfuse.callback import CallbackHandler  # type: ignore

        public_key = os.getenv("LANGFUSE_PUBLIC_KEY", None)
        secret_key = os.getenv("LANGFUSE_SECRET_KEY", None)
        if public_key is None or secret_key is None:
            raise ValueError("LANGFUSE_PUBLIC_KEY or LANGFUSE_SECRET_KEY is not set")

        callback = CallbackHandler(
            public_key=public_key,
            secret_key=secret_key,
            host=config.tracing.langfuse.host,
        )
        callbacks.append(callback)

    if config.tracing.langsmith.use_langsmith or override_use_langsmith:
        os.environ["LANGCHAIN_TRACING_V2"] = "true"
        os.environ["LANGCHAIN_PROJECT"] = config.tracing.project
        api_key = os.getenv("LANGCHAIN_API_KEY", None)
        if api_key is None:
            raise ValueError("LANGCHAIN_API_KEY is not set")
        callback = LangChainTracer(
            project_name=config.tracing.project,
            client=Client(api_key=api_key, api_url=config.tracing.langsmith.host),
        )
        callbacks.append(callback)

    return callbacks



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Literal, Type

import numpy as np
from geometry_msgs.msg import Point, Pose, PoseStamped, Quaternion
from pydantic import BaseModel, Field
from tf2_geometry_msgs import do_transform_pose

from rai.communication.ros2.ros_async import get_future_result
from rai.tools.ros2.base import BaseROS2Tool

try:
    from rai_interfaces.srv import ManipulatorMoveTo
except ImportError:
    logging.warning(
        "rai_interfaces is not installed, ManipulatorMoveTo tool will not work."
    )

try:
    from rai_open_set_vision.tools import GetGrabbingPointTool
except ImportError:
    logging.warning(
        "rai_open_set_vision is not installed, GetGrabbingPointTool will not work"
    )


class MoveToPointToolInput(BaseModel):
    x: float = Field(description="The x coordinate of the point to move to")
    y: float = Field(description="The y coordinate of the point to move to")
    z: float = Field(description="The z coordinate of the point to move to")
    task: Literal["grab", "drop"] = Field(
        description="Specify the intended action: use 'grab' to pick up an object, or 'drop' to release it. "
        "This determines the gripper's behavior during the movement."
    )


class MoveToPointTool(BaseROS2Tool):
    name: str = "move_to_point"
    description: str = (
        "Guide the robot's end effector to a specific point within the manipulator's operational space. "
        "This tool ensures precise movement to the desired location. "
        "While it confirms successful positioning, please note that it doesn't provide feedback on the "
        "success of grabbing or releasing objects. Use additional sensors or tools for that information."
    )

    manipulator_frame: str = Field(..., description="Manipulator frame")
    min_z: float = Field(default=0.135, description="Minimum z coordinate [m]")
    calibration_x: float = Field(default=0.0, description="Calibration x [m]")
    calibration_y: float = Field(default=0.0, description="Calibration y [m]")
    calibration_z: float = Field(default=0.0, description="Calibration z [m]")
    additional_height: float = Field(
        default=0.05, description="Additional height for the place task [m]"
    )

    # constant quaternion
    quaternion: Quaternion = Field(
        default=Quaternion(x=0.9238795325112867, y=-0.3826834323650898, z=0.0, w=0.0),
        description="Constant quaternion",
    )

    args_schema: Type[MoveToPointToolInput] = MoveToPointToolInput

    def _run(
        self,
        x: float,
        y: float,
        z: float,
        task: Literal["grab", "drop"],
    ) -> str:
        client = self.connector.node.create_client(
            ManipulatorMoveTo,
            "/manipulator_move_to",
        )
        pose_stamped = PoseStamped()
        pose_stamped.header.frame_id = self.manipulator_frame
        pose_stamped.pose = Pose(
            position=Point(x=x, y=y, z=z),
            orientation=self.quaternion,
        )

        if task == "drop":
            pose_stamped.pose.position.z += self.additional_height

        pose_stamped.pose.position.x += self.calibration_x
        pose_stamped.pose.position.y += self.calibration_y
        pose_stamped.pose.position.z += self.calibration_z

        pose_stamped.pose.position.z = np.max(
            [pose_stamped.pose.position.z, self.min_z]
        )

        request = ManipulatorMoveTo.Request()
        request.target_pose = pose_stamped

        if task == "grab":
            request.initial_gripper_state = True  # open
            request.final_gripper_state = False  # closed
        else:
            request.initial_gripper_state = False  # closed
            request.final_gripper_state = True  # open

        future = client.call_async(request)
        self.connector.node.get_logger().debug(
            f"Calling ManipulatorMoveTo service with request: x={request.target_pose.pose.position.x:.2f}, y={request.target_pose.pose.position.y:.2f}, z={request.target_pose.pose.position.z:.2f}"
        )
        response = get_future_result(future, timeout_sec=5.0)
        if response is None:
            return f"Service call failed for point ({x:.2f}, {y:.2f}, {z:.2f})."

        if response.success:
            return f"End effector successfully positioned at coordinates ({x:.2f}, {y:.2f}, {z:.2f}). Note: The status of object interaction (grab/drop) is not confirmed by this movement."
        else:
            return f"Failed to position end effector at coordinates ({x:.2f}, {y:.2f}, {z:.2f})."


class GetObjectPositionsToolInput(BaseModel):
    object_name: str = Field(
        ..., description="The name of the object to get the positions of"
    )


class GetObjectPositionsTool(BaseROS2Tool):
    name: str = "get_object_positions"
    description: str = (
        "Retrieve the positions of all objects of a specified type in the target frame. "
        "This tool provides accurate positional data but does not distinguish between different colors of the same object type. "
        "While position detection is reliable, please note that object classification may occasionally be inaccurate."
    )

    target_frame: str
    source_frame: str
    camera_topic: str  # rgb camera topic
    depth_topic: str
    camera_info_topic: str  # rgb camera info topic
    get_grabbing_point_tool: "GetGrabbingPointTool"

    args_schema: Type[GetObjectPositionsToolInput] = GetObjectPositionsToolInput

    @staticmethod
    def format_pose(pose: Pose):
        return f"Centroid(x={pose.position.x:.2f}, y={pose.position.y:2f}, z={pose.position.z:2f})"

    def _run(self, object_name: str):
        transform = self.connector.get_transform(
            target_frame=self.target_frame, source_frame=self.source_frame
        )

        results = self.get_grabbing_point_tool._run(
            camera_topic=self.camera_topic,
            depth_topic=self.depth_topic,
            camera_info_topic=self.camera_info_topic,
            object_name=object_name,
        )

        poses = []
        for result in results:
            cam_pose = result[0]
            poses.append(
                Pose(position=Point(x=cam_pose[0], y=cam_pose[1], z=cam_pose[2]))
            )

        mani_frame_poses = []
        for pose in poses:
            mani_frame_pose = do_transform_pose(pose, transform)
            mani_frame_poses.append(mani_frame_pose)

        if len(mani_frame_poses) == 0:
            return f"No {object_name}s detected."
        else:
            return f"Centroids of detected {object_name}s in {self.target_frame} frame: [{', '.join(map(self.format_pose, mani_frame_poses))}]. Sizes of the detected objects are unknown."



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Type,
    TypeVar,
    get_args,
)
from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field


class ConnectorException(Exception):
    """Base exception for all connector exceptions."""

    pass


class BaseMessage(BaseModel):
    payload: Any = Field(
        default=None,
        description="Payload is meant for non-validated data if such data is present.",
    )
    metadata: Dict[str, Any] = Field(default_factory=dict)
    timestamp: float = Field(default_factory=time.time)

    model_config = ConfigDict(arbitrary_types_allowed=True)


T = TypeVar("T", bound=BaseMessage)


class ParametrizedCallback(BaseModel, Generic[T]):
    # Callback is of type T if raw is False, otherwise it is of type Any
    callback: Callable[[T | Any], None]
    raw: bool
    id: str = Field(default_factory=lambda: str(uuid4()))


class BaseConnector(Generic[T]):
    def __init__(self, callback_max_workers: int = 4):
        self.callback_max_workers = callback_max_workers
        self.logger = logging.getLogger(self.__class__.__name__)
        self.registered_callbacks: Dict[str, Dict[str, ParametrizedCallback[T]]] = (
            defaultdict(dict)
        )
        self.callback_id_mapping: Dict[str, tuple[str, ParametrizedCallback[T]]] = {}
        self.callback_executor = ThreadPoolExecutor(
            max_workers=self.callback_max_workers
        )

        if not hasattr(self, "__orig_bases__"):
            self.__orig_bases__ = {}
            raise ConnectorException(
                f"Error while instantiating {str(self.__class__)}: "
                "Message type T derived from BaseMessage needs to be provided"
                " e.g. Connector[MessageType]()"
            )
        self.T_class: Type[T] = get_args(self.__orig_bases__[-1])[0]

    def _generate_handle(self) -> str:
        return str(uuid4())

    def send_message(self, message: T, target: str, **kwargs: Optional[Any]) -> None:
        """Implements publish pattern.

        Sends a message to one or more subscribers. The target parameter
        can be used to specify the destination or topic.

        Raises:
            ConnectorException: If the message cannot be sent.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def receive_message(
        self, source: str, timeout_sec: float, **kwargs: Optional[Any]
    ) -> T:
        """Implements subscribe pattern.

        Receives a message from a publisher. The source parameter
        can be used to specify the source or topic to subscribe to.

        Raises:
            ConnectorException: If the message cannot be received.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def register_callback(
        self,
        source: str,
        callback: Callable[[T | Any], None],
        raw: bool = False,
        **kwargs: Optional[Any],
    ) -> str:
        """Implements register callback.

        Registers a callback to be called when a message is received from a source.
        If raw is False, the callback will receive a T object.
        If raw is True, the callback will receive the raw message.

        Raises:
            ConnectorException: If the callback cannot be registered.
        """
        parametrized_callback = ParametrizedCallback[T](callback=callback, raw=raw)
        self.registered_callbacks[source][parametrized_callback.id] = (
            parametrized_callback
        )
        self.callback_id_mapping[parametrized_callback.id] = (
            source,
            parametrized_callback,
        )
        return parametrized_callback.id

    def unregister_callback(self, callback_id: str) -> None:
        """Unregisters a callback from a source.

        Args:
            callback_id: The id of the callback to unregister.

        Raises:
            ConnectorException: If the callback cannot be unregistered.
        """
        if callback_id not in self.callback_id_mapping:
            raise ConnectorException(f"Callback with id {callback_id} not found.")

        source, _ = self.callback_id_mapping[callback_id]
        del self.registered_callbacks[source][callback_id]
        del self.callback_id_mapping[callback_id]

    def _safe_callback_wrapper(self, callback: Callable[[T], None], message: T) -> None:
        """Safely execute a callback with error handling.

        Args:
            callback: The callback function to execute
            message: The message to pass to the callback
        """
        try:
            callback(message)
        except Exception as e:
            self.logger.error(f"Error in callback: {str(e)}")

    def general_callback(self, source: str, message: Any) -> None:
        """General callback for all messages.
        Use through functools.partial to pass source."""
        processed_message = self.general_callback_preprocessor(message)
        for parametrized_callback in self.registered_callbacks.get(source, {}).values():
            payload = message if parametrized_callback.raw else processed_message
            self.callback_executor.submit(
                self._safe_callback_wrapper, parametrized_callback.callback, payload
            )

    def general_callback_preprocessor(self, message: Any) -> T:
        """Preprocessor for general callback used to transform any message to a BaseMessage."""
        raise NotImplementedError("This method should be implemented by the subclass.")

    def service_call(
        self, message: T, target: str, timeout_sec: float, **kwargs: Optional[Any]
    ) -> BaseMessage:
        """Implements request-response pattern.

        Sends a request and waits for a response. The target parameter
        specifies the service endpoint to call.

        Raises:
            ConnectorException: If the service call cannot be made.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def create_service(
        self,
        service_name: str,
        on_request: Callable,
        on_done: Optional[Callable] = None,
        **kwargs: Optional[Any],
    ) -> str:
        """Sets up a service endpoint for handling requests.

        Creates a service that can receive and process requests.
        The on_request callback handles incoming requests,
        and on_done (if provided) is called when the service is terminated.

        Raises:
            ConnectorException: If the service cannot be created.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def create_action(
        self,
        action_name: str,
        generate_feedback_callback: Callable,
        **kwargs: Optional[Any],
    ) -> str:
        """Sets up an action endpoint for long-running operations.

        Creates an action that can be started and monitored.
        The generate_feedback_callback is used to provide progress updates
        during the action's execution.

        Raises:
            ConnectorException: If the action cannot be created.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def start_action(
        self,
        action_data: Optional[T],
        target: str,
        on_feedback: Callable,
        on_done: Callable,
        timeout_sec: float,
        **kwargs: Optional[Any],
    ) -> str:
        """Initiates a long-running operation with feedback.

        Starts an action and provides callbacks for feedback and completion.
        The on_feedback callback receives progress updates,
        and on_done is called when the action completes.

        Raises:
            ConnectorException: If the action cannot be started.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def terminate_action(self, action_handle: str, **kwargs: Optional[Any]) -> Any:
        """Cancels an ongoing action.

        Stops the execution of a previously started action.
        The action_handle identifies which action to terminate.

        Raises:
            ConnectorException: If the action cannot be terminated.
        """
        raise NotImplementedError("This method should be implemented by the subclass.")

    def shutdown(self):
        """Shuts down the connector and releases all resources."""
        self.callback_executor.shutdown(wait=True)



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import inspect
from typing import Any, Callable, Dict, List, TypeVar

import cv2
import numpy as np
import streamlit as st
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.messages import BaseMessage
from langchain_core.runnables import Runnable, RunnableConfig
from streamlit.delta_generator import DeltaGenerator
from streamlit.runtime.scriptrunner import add_script_run_ctx, get_script_run_ctx

# code inspired by (mostly copied, some changes were applied and might be updated in the future)
# https://github.com/shiv248/Streamlit-x-LangGraph-Cookbooks/tree/b8e623bdc9821fc1cf581607454dae1afc054df2/tool_calling_via_callback


# Define a function to create a callback handler for Streamlit that updates the UI dynamically
def get_streamlit_cb(parent_container: DeltaGenerator) -> BaseCallbackHandler:
    """
    Creates a Streamlit callback handler that updates the provided Streamlit container with new tokens.
    Args:
        parent_container (DeltaGenerator): The Streamlit container where the text will be rendered.
    Returns:
        BaseCallbackHandler: An instance of a callback handler configured for Streamlit.
    """

    # Define a custom callback handler class for managing and displaying stream events in Streamlit
    class StreamHandler(BaseCallbackHandler):
        """
        Custom callback handler for Streamlit that updates a Streamlit container with new tokens.
        """

        def __init__(
            self, container: st.delta_generator.DeltaGenerator, initial_text: str = ""
        ):
            """
            Initializes the StreamHandler with a Streamlit container and optional initial text.
            Args:
                container (st.delta_generator.DeltaGenerator): The Streamlit container where text will be rendered.
                initial_text (str): Optional initial text to start with in the container.
            """
            self.container = container  # The Streamlit container to update
            self.thoughts_placeholder = (
                self.container.container()
            )  # container to hold tool_call renders
            self.tool_output_placeholder = None  # placeholder for the output of the tool call to be in the expander
            self.token_placeholder = self.container.empty()  # for token streaming
            self.text = (
                initial_text  # The text content to display, starting with initial text
            )

        def on_llm_new_token(self, token: str, **kwargs) -> None:
            """
            Callback method triggered when a new token is received (e.g., from a language model).
            Args:
                token (str): The new token received.
                **kwargs: Additional keyword arguments.
            """
            self.text += token  # Append the new token to the existing text
            self.token_placeholder.write(self.text)

        def on_tool_start(
            self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
        ) -> None:
            """
            Run when the tool starts running.
            Args:
                serialized (Dict[str, Any]): The serialized tool.
                input_str (str): The input string.
                kwargs (Any): Additional keyword arguments.
            """
            with self.thoughts_placeholder:
                status_placeholder = st.empty()  # Placeholder to show the tool's status
                with status_placeholder.status("Calling Tool...", expanded=True) as s:
                    st.write("Called tool: ", serialized["name"])
                    st.write("tool description: ", serialized["description"])
                    st.write("tool input: ")
                    st.code(input_str)  # Display the input data sent to the tool
                    st.write("tool output: ")
                    # Placeholder for tool output that will be updated later below
                    self.tool_output_placeholder = st.empty()
                    s.update(
                        label="Completed Calling Tool!", expanded=False
                    )  # Update the status once done

        def on_tool_end(self, output: Any, **kwargs: Any) -> Any:
            """
            Run when the tool ends.
            Args:
                output (Any): The output from the tool.
                kwargs (Any): Additional keyword arguments.
            """
            # We assume that `on_tool_end` comes after `on_tool_start`, meaning output_placeholder exists
            if self.tool_output_placeholder:
                with self.tool_output_placeholder.container():
                    st.code(output.content)  # Display the tool's output

                    if output.artifact is not None:
                        if "images" in output.artifact:  # Display available images
                            for image in output.artifact["images"]:
                                image_cv2 = cv2.imdecode(
                                    np.frombuffer(base64.b64decode(image), np.uint8),
                                    cv2.IMREAD_COLOR,
                                )
                                st.image(
                                    image_cv2, channels="BGR", use_container_width=True
                                )

    # Define a type variable for generic type hinting in the decorator, to maintain
    # input function and wrapped function return type
    fn_return_type = TypeVar("fn_return_type")

    # Decorator function to add the Streamlit execution context to a function
    def add_streamlit_context(
        fn: Callable[..., fn_return_type],
    ) -> Callable[..., fn_return_type]:
        """
        Decorator to ensure that the decorated function runs within the Streamlit execution context.
        Args:
            fn (Callable[..., fn_return_type]): The function to be decorated.
        Returns:
            Callable[..., fn_return_type]: The decorated function that includes the Streamlit context setup.
        """
        ctx = (
            get_script_run_ctx()
        )  # Retrieve the current Streamlit script execution context

        def wrapper(*args, **kwargs) -> fn_return_type:
            """
            Wrapper function that adds the Streamlit context and then calls the original function.
            Args:
                *args: Positional arguments to pass to the original function.
                **kwargs: Keyword arguments to pass to the original function.
            Returns:
                fn_return_type: The result from the original function.
            """
            add_script_run_ctx(
                ctx=ctx
            )  # Add the Streamlit context to the current execution
            return fn(*args, **kwargs)  # Call the original function with its arguments

        return wrapper

    # Create an instance of the custom StreamHandler with the provided Streamlit container
    st_cb = StreamHandler(parent_container)

    # Iterate over all methods of the StreamHandler instance
    for method_name, method_func in inspect.getmembers(
        st_cb, predicate=inspect.ismethod
    ):
        if method_name.startswith("on_"):  # Identify callback methods
            setattr(
                st_cb, method_name, add_streamlit_context(method_func)
            )  # Wrap and replace the method

    # Return the fully configured StreamHandler instance with the context-aware callback methods
    return st_cb


def streamlit_invoke(
    graph: Runnable[Any, Any], messages: List[BaseMessage], callables: List[Callable]
):
    if not isinstance(callables, list):
        raise TypeError("callables must be a list")
    return graph.invoke(
        {"messages": messages},
        config=RunnableConfig({"callbacks": callables, "recursion_limit": 100}),
    )



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from subprocess import PIPE, Popen
from threading import Timer
from typing import List, Literal, Optional

from langchain_core.tools import BaseTool, BaseToolkit, tool

FORBIDDEN_CHARACTERS = ["&", ";", "|", "&&", "||", "(", ")", "<", ">", ">>", "<<"]


def run_with_timeout(cmd: List[str], timeout_sec: int):
    proc = Popen(cmd, stdout=PIPE, stderr=PIPE)
    timer = Timer(timeout_sec, proc.kill)
    try:
        timer.start()
        stdout, stderr = proc.communicate()
        return stdout, stderr
    finally:
        timer.cancel()


def run_command(cmd: List[str], timeout: int = 5):
    # Validate command safety by checking for shell operators
    # Block potentially dangerous characters
    if any(char in " ".join(cmd) for char in FORBIDDEN_CHARACTERS):
        raise ValueError(
            "Command is not safe to run. The command contains forbidden characters."
        )
    stdout, stderr = run_with_timeout(cmd, timeout)
    output = {}
    if stdout:
        output["stdout"] = stdout.decode("utf-8")
    else:
        output["stdout"] = "Command returned no stdout output"
    if stderr:
        output["stderr"] = stderr.decode("utf-8")
    else:
        output["stderr"] = "Command returned no stderr output"
    return str(output)


class ROS2CLIToolkit(BaseToolkit):
    def get_tools(self) -> List[BaseTool]:
        return [
            ros2_action,
            ros2_service,
            ros2_node,
            ros2_param,
            ros2_interface,
            ros2_topic,
        ]


@tool
def ros2_action(
    command: Literal["info", "list", "type", "send_goal"],
    arguments: Optional[List[str]] = None,
    timeout: int = 5,
):
    """Run a ROS2 action command
    Args:
        command: The action command to run (info/list/type)
        arguments: Additional arguments for the command as a list of strings
        timeout: Command timeout in seconds
    """
    cmd = ["ros2", "action", command]
    if arguments:
        cmd.extend(arguments)
    return run_command(cmd, timeout)


@tool
def ros2_service(
    command: Literal["call", "find", "info", "list", "type"],
    arguments: Optional[List[str]] = None,
    timeout: int = 5,
):
    """Run a ROS2 service command
    Args:
        command: The service command to run
        arguments: Additional arguments for the command as a list of strings
        timeout: Command timeout in seconds
    """
    cmd = ["ros2", "service", command]
    if arguments:
        cmd.extend(arguments)
    return run_command(cmd, timeout)


@tool
def ros2_node(
    command: Literal["info", "list"],
    arguments: Optional[List[str]] = None,
    timeout: int = 5,
):
    """Run a ROS2 node command
    Args:
        command: The node command to run
        arguments: Additional arguments for the command as a list of strings
        timeout: Command timeout in seconds
    """
    cmd = ["ros2", "node", command]
    if arguments:
        cmd.extend(arguments)
    return run_command(cmd, timeout)


@tool
def ros2_param(
    command: Literal["delete", "describe", "dump", "get", "list", "set"],
    arguments: Optional[List[str]] = None,
    timeout: int = 5,
):
    """Run a ROS2 parameter command
    Args:
        command: The parameter command to run
        arguments: Additional arguments for the command as a list of strings
        timeout: Command timeout in seconds
    """
    cmd = ["ros2", "param", command]
    if arguments:
        cmd.extend(arguments)
    return run_command(cmd, timeout)


@tool
def ros2_interface(
    command: Literal["list", "package", "packages", "proto", "show"],
    arguments: Optional[List[str]] = None,
    timeout: int = 5,
):
    """Run a ROS2 interface command
    Args:
        command: The interface command to run
        arguments: Additional arguments for the command as a list of strings
        timeout: Command timeout in seconds
    """
    cmd = ["ros2", "interface", command]
    if arguments:
        cmd.extend(arguments)
    return run_command(cmd, timeout)


@tool
def ros2_topic(
    command: Literal[
        "bw", "delay", "echo", "find", "hz", "info", "list", "pub", "type"
    ],
    arguments: Optional[List[str]] = None,
    timeout: int = 5,
):
    """Run a ROS2 topic command
    Args:
        command: The topic command to run:
            - bw: Display bandwidth used by topic
            - delay: Display delay of topic from timestamp in header
            - echo: Output messages from a topic
            - find: Output a list of available topics of a given type
            - hz: Print the average publishing rate to screen
            - info: Print information about a topic
            - list: Output a list of available topics
            - pub: Publish a message to a topic
            - type: Print a topic's type
        arguments: Additional arguments for the command as a list of strings
        timeout: Command timeout in seconds
    """
    cmd = ["ros2", "topic", command]
    if arguments:
        cmd.extend(arguments)
    return run_command(cmd, timeout)



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated, List, Optional

from langchain_core.tools import BaseTool, BaseToolkit
from pydantic import ConfigDict

from rai.communication.ros2.connectors import ROS2Connector


class BaseROS2Tool(BaseTool):
    """
    Base class for all ROS2 tools.

    Parameters
    ----------
    connector : ROS2Connector
        The connector to the ROS 2 system.
    readable : Optional[List[str]]
        The topics that can be read. If the list is not provided, all topics can be read.
    writable : Optional[List[str]]
        The names (topics/actions/services) that can be written. If the list is not provided, all topics can be written.
    forbidden : Optional[List[str]]
        The names (topics/actions/services) that are forbidden to read and write.
    """

    connector: ROS2Connector
    readable: Optional[List[str]] = None
    writable: Optional[List[str]] = None
    forbidden: Optional[List[str]] = None

    name: str = ""
    description: str = ""

    def is_readable(self, topic: str) -> bool:
        if self.forbidden is not None and topic in self.forbidden:
            return False
        if self.readable is None:
            return True
        return topic in self.readable

    def is_writable(self, topic: str) -> bool:
        if self.forbidden is not None and topic in self.forbidden:
            return False
        if self.writable is None:
            return True
        return topic in self.writable


class BaseROS2Toolkit(BaseToolkit):
    """
    Base class for all ROS2 toolkits.

    Parameters
    ----------
    connector : ROS2Connector
        The connector to the ROS2 system.
    readable : Optional[List[str]]
        The topics that can be read. If the list is not provided, all topics can be read.
    writable : Optional[List[str]]
        The names (topics/actions/services) that can be written. If the list is not provided, all topics can be written.
    forbidden : Optional[List[str]]
        The names (topics/actions/services) that are forbidden to read and write.
    """

    connector: ROS2Connector
    readable: Optional[
        Annotated[
            List[str],
            """The topics that can be read.
            If the list is not provided, all topics can be read.""",
        ]
    ] = None
    writable: Optional[
        Annotated[
            List[str],
            """The names (topics/actions/services) that can be written.
            If the list is not provided, all topics can be written.""",
        ]
    ] = None
    forbidden: Optional[
        Annotated[
            List[str],
            """The names (topics/actions/services) that are forbidden to read and write.""",
        ]
    ] = None

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import operator
from typing import Annotated, List, Type, TypedDict

from langchain.agents.agent import AgentExecutor
from langchain.agents.tool_calling_agent.base import create_tool_calling_agent
from langchain_community.vectorstores import FAISS
from langchain_core.documents import Document
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage
from langchain_core.prompts.chat import ChatPromptTemplate
from langchain_core.tools import BaseTool
from langchain_core.vectorstores import VectorStore
from langgraph.graph import StateGraph
from pydantic import BaseModel, Field

from rai.apps.document_loader import ingest_documentation
from rai.initialization import get_embeddings_model

logging.basicConfig(level=logging.WARN)


class QueryDocsToolInput(BaseModel):
    query: str = Field(
        ..., description="The query string to be searched in the documentation."
    )


class QueryDocsTool(BaseTool):
    name: str = "query_docs"
    description: str = "Query the similarity search database."

    args_schema: Type[QueryDocsToolInput] = QueryDocsToolInput

    vector_store: VectorStore = Field(
        ..., description="The vector store to be searched."
    )
    k: int = Field(3, description="The number of results to return.")

    def _run(self, query: str) -> List[Document]:
        return self.vector_store.similarity_search(query, k=self.k)


class State(TypedDict):
    messages: Annotated[list[BaseMessage], operator.add]


def talk_to_docs(documentation_root: str, llm: BaseChatModel):
    docs = ingest_documentation(documentation_root)
    vector_store = FAISS.from_documents(docs, get_embeddings_model())

    query_docs = QueryDocsTool(vector_store=vector_store, k=3)

    prompt = ChatPromptTemplate.from_template(
        "You are a robot called with access to your documentation. "
        "You use a lot of emojis when responding to users. "
        "You always respond in first person. "
        "Your main goal is to help the user find the information they need."
        "Your access to the documentation is not important to the user. "
        "You do not have to inform the user about it unless you did not find needed information."
        "Here is the conversation history: {messages} {agent_scratchpad}"
    )

    agent = create_tool_calling_agent(llm, [query_docs], prompt)  # type: ignore
    agent_executor = AgentExecutor(
        agent=agent,
        tools=[query_docs],
        return_intermediate_steps=True,  # type: ignore
    )

    def input_node(state: State) -> State:
        user_message = HumanMessage(content=input("You: "))
        return {"messages": [user_message]}

    def agent_node(state: State) -> State:
        messages = state["messages"]
        response = agent_executor.invoke({"messages": messages})
        return {"messages": [AIMessage(content=response["output"])]}

    workflow = StateGraph(State)
    workflow.add_node("user_input", input_node)
    workflow.add_node("agent", agent_node)
    workflow.add_edge("user_input", "agent")
    workflow.add_edge("agent", "user_input")
    workflow.set_entry_point("user_input")

    graph = workflow.compile()
    return graph



# Copyright (C) 2024 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import time
import uuid
from typing import Callable, List, Optional

from langchain.agents import AgentExecutor, create_tool_calling_agent
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import (
    BaseMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnableConfig
from langchain_core.tools import BaseTool, tool
from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)

TASK_EXECUTOR_PROMPT = """
I am a robot's execution task system. I am responsible for executing the current task using reasoning and available tooling. I am focused on the task's completion.
When the simplest solution is not enough, I will reason about the task and use available tools to complete it.
My job is complete whenever the task is finished or the task is deemed impossible to complete.
Sometimes my job can be interrupted, and I will need to resume it later. I will be given a previous reasoning and state history to help me continue where I left off.
"""

TASK_SUMMARIZER_PROMPT = """
I am a robot's task summarizer. I am responsible for summarizing the task's completion status.
"""


@tool
def finish_execution(bool: bool) -> str:
    """This tool is used to stop execution when the task is considered done or unrecoverable."""
    return "The task is finished."


class TaskState(BaseModel):
    done: bool = Field(
        ..., description="True if the robot has reached the final state of the task."
    )
    summary: str = Field(
        ..., description="Message describing the robot's mission fulfilness."
    )


class Task(BaseModel):
    task: str = Field(..., description="The task to be executed by the robot.")


def task_executor(
    llm: BaseChatModel,
    task: Task,
    tools: List[BaseTool],
    task_executor_prompt: str = TASK_EXECUTOR_PROMPT,
    task_summarizer_prompt: str = TASK_SUMMARIZER_PROMPT,
) -> TaskState:
    agent_template = ChatPromptTemplate.from_messages(
        [
            ("system", task_executor_prompt),
            ("human", "The task: {task}"),
            ("placeholder", "{agent_scratchpad}"),
        ]
    )
    agent = create_tool_calling_agent(llm=llm, tools=tools, prompt=agent_template)
    agent_executor = AgentExecutor(
        agent=agent, tools=tools, verbose=True, return_intermediate_steps=True
    )

    agent_output = agent_executor.invoke({"task": task.task})

    agent_reasoning = f"intermediate_steps: {str(agent_output['intermediate_steps'])} final_output: {agent_output['output']}"
    history = [
        SystemMessage(content=task_summarizer_prompt),
        HumanMessage(content=agent_reasoning, name="agent"),
    ]
    task_state = llm.with_structured_output(schema=TaskState).invoke(history)

    return task_state


def structured_task_executor(
    llm: BaseChatModel,
    task: Task,
    tools: List[BaseTool],
    task_executor_prompt: str = TASK_EXECUTOR_PROMPT,
    task_summarizer_prompt: str = TASK_SUMMARIZER_PROMPT,
    callbacks: Optional[List[Callable[[], str]]] = None,
    rate: float = 1,
    max_iters: int = 10,
) -> TaskState:
    def get_latest_data():
        if callbacks is None:
            return ""
        return "\n".join([callback() for callback in callbacks])

    invoke_config = RunnableConfig(metadata={"task": task.task, "run_id": uuid.uuid4()})
    task_history: List[BaseMessage] = [
        SystemMessage(content=task_executor_prompt),
        HumanMessage(content=f"The task: {task.task}"),
    ]
    tools = tools + [finish_execution]

    tool_name_to_tool_map = {tool.name: tool for tool in tools}

    llm_with_tools = llm.bind_tools(tools)

    it = 0
    logger.info("Entering the task execution loop.")
    while it < max_iters:
        logger.info(f"Task execution loop iteration: {it + 1}/{max_iters}")
        it += 1
        loop_start = time.time()
        state = get_latest_data()
        task_history += [HumanMessage(content=state, name="state")]
        output = llm_with_tools.invoke(task_history, config=invoke_config)

        tools_outputs: List[ToolMessage] = []
        if output.tool_calls:
            for tool_call in output.tool_calls:
                tool_name = tool_call["name"]
                tool_instance = tool_name_to_tool_map[tool_name]
                tool_output = tool_instance.run(tool_call["args"])
                tools_outputs.append(
                    ToolMessage(tool_call_id=tool_call["id"], content=tool_output)
                )

        task_history += [output]
        task_history += tools_outputs
        loop_end = time.time()

        time.sleep(max(0, 1 / rate - (loop_end - loop_start)))
        if "finish_execution" in [tool["name"] for tool in output.tool_calls]:
            logging.info("Finishing the task execution. Reason: finish requested.")
            break
    logger.info("Exiting the task execution loop.")
    history = [
        SystemMessage(content=task_summarizer_prompt),
        *task_history[1:],  # skip the first system message
    ]
    task_state = llm.with_structured_output(schema=TaskState).invoke(history)

    return task_state



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict, List, Optional, Tuple, Union

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage, HumanMessage
from langchain_core.tools import BaseTool
from pydantic import BaseModel, ConfigDict, Field

from rai.agents.langchain.agent import LangChainAgent, newMessageBehaviorType
from rai.agents.langchain.runnables import ReActAgentState, create_state_based_runnable
from rai.aggregators import BaseAggregator
from rai.communication.base_connector import BaseConnector
from rai.communication.hri_connector import HRIConnector, HRIMessage
from rai.messages.multimodal import HumanMultimodalMessage


class StateBasedConfig(BaseModel):
    aggregators: Dict[Union[str, Tuple[str, str]], List[BaseAggregator]] = Field(
        description="Dict of topic : aggregator or (topic, msg_type) : aggragator"
    )
    time_interval: float = Field(default=5.0)
    max_workers: int = 8

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


class BaseStateBasedAgent(LangChainAgent, ABC):
    """
    Agent that runs aggregators (config.aggregators) every config.time_interval seconds.
    Aggregators are registered to their sources using
    :py:class:`~rai.communication.ros2.connectors.ROS2Connector`

    Output from aggragators is called `state`. Such state is saved and can be
    retrieved by `get_state` method.

    In `StateBaseAgent`, state is added to LLM history. For more details about the LLM
    agent see :py:func:`~rai.agents.langchain.runnables.create_state_based_runnable`
    """

    def __init__(
        self,
        config: StateBasedConfig,
        target_connectors: Dict[str, HRIConnector[HRIMessage]],
        llm: Optional[BaseChatModel] = None,
        tools: Optional[List[BaseTool]] = None,
        state: Optional[ReActAgentState] = None,
        system_prompt: Optional[str] = None,
        new_message_behavior: newMessageBehaviorType = "interrupt_keep_last",
        max_size: int = 100,
    ):
        runnable = create_state_based_runnable(
            llm=llm,
            tools=tools,
            system_prompt=system_prompt,
            state_retriever=self.get_state,
        )
        super().__init__(
            target_connectors=target_connectors,
            runnable=runnable,
            state=state,
            new_message_behavior=new_message_behavior,
            max_size=max_size,
        )
        self.config = config

        self._aggregation_results: Dict[str, HumanMessage | HumanMultimodalMessage] = (
            dict()
        )
        self._aggregation_thread: threading.Thread | None = None

        self._registered_callbacks = set()
        self._connector = self.setup_connector()
        self._configure_state_sources()

    @abstractmethod
    def setup_connector(self) -> BaseConnector:
        pass

    def _configure_state_sources(self):
        for source, aggregators in self.config.aggregators.items():
            if isinstance(source, tuple):
                source, msg_type = source
            else:
                msg_type = None
            for aggregator in aggregators:
                callback_id = self._connector.register_callback(
                    source, aggregator, raw=True, msg_type=msg_type
                )
                self._registered_callbacks.add(callback_id)

    def run(self):
        super().run()
        self._aggregation_thread = threading.Thread(target=self._run_state_loop)
        self._aggregation_thread.start()

    def get_state(self) -> Dict[str, HumanMessage | HumanMultimodalMessage]:
        """Returns output for all aggregators"""
        return self._aggregation_results

    def _run_state_loop(self):
        """Runs aggregation on collected data"""
        while not self._stop_event.is_set():
            ts = time.perf_counter()
            self.logger.debug("Starting aggregation interval")
            self._on_aggregation_interval()
            elapsed_time = time.perf_counter() - ts
            self.logger.debug(f"Aggregation done in: {elapsed_time:.2f}s")
            if elapsed_time > self.config.time_interval:
                self.logger.warning(
                    "State aggregation time interval exceeded. Expected "
                    f"{self.config.time_interval:.2f}s, got {elapsed_time:.2f}s. Consider "
                    f"increasing {self.__class__.__name__}.config.time_interval."
                )
            time.sleep(max(0, self.config.time_interval - (elapsed_time)))

    def _on_aggregation_interval(self):
        """Runs aggregation on collected data"""

        def process_aggregator(
            source: str, aggregator: BaseAggregator
        ) -> Tuple[str, BaseMessage | None]:
            self.logger.info(
                f"Running aggregator: {aggregator}(source={source}) on {len(aggregator.get_buffer())} messages"
            )
            ts = time.perf_counter()

            output = aggregator.get()

            self.logger.debug(
                f'Aggregator "{aggregator}(source={source})" done in {time.perf_counter() - ts:.2f}s'
            )
            return source, output

        with ThreadPoolExecutor(max_workers=self.config.max_workers) as executor:
            futures = list()
            for source, aggregators in self.config.aggregators.items():
                for aggregator in aggregators:
                    future = executor.submit(process_aggregator, source, aggregator)
                    futures.append(future)

            for future in as_completed(futures):
                try:
                    source, output = future.result()
                except Exception as e:
                    self.logger.error(f"Aggregator crashed: {e}")
                    continue

                if output is None:
                    continue
                self._aggregation_results[source] = output

    def stop(self):
        """Stop the agent's execution loop."""
        self._stop_event.set()
        self._interrupt_event.set()
        self._agent_ready_event.wait()
        if self._thread is not None:
            self.logger.info("Stopping the agent. Please wait...")
            self._thread.join()
            self._thread = None
        if self._aggregation_thread is not None:
            self._aggregation_thread.join()
            self._aggregation_thread = None
        for callback_id in self._registered_callbacks:
            self._connector.unregister_callback(callback_id)
        self._stop_event.clear()
        self._connector.shutdown()
        self.logger.info("Agent stopped")
